# -*- coding: utf-8 -*-
# Distributed under the MIT License.
# Copied and modified from the porespy project.
# See https://github.com/PMEAL/porespy/blob/dev/porespy/generators/_imgen.py for more information.
# See https://github.com/PMEAL/porespy/blob/dev/LICENSE for the license.

from scipy.ndimage import distance_transform_edt as edt
import numpy as np
from pumapy.utilities.workspace import Workspace
from pumapy.materialproperties.volumefraction import compute_volume_fraction
import sys


def generate_random_fibers(shape, radius, nfibers=None, porosity=None, phi=0, theta=90, length=None, max_iter=3):
    """ Generates random fibers from number of fibers or porosity

    :param shape: the size of the workspace to generate in (Nx, Ny, Nz) where N is the number of voxels.
    :type shape: tuple(int, int, int)
    :param radius: the radius of the fibers in voxels
    :type radius: int
    :param nfibers: the number of fibers to add to the domain. Adjust this value to control the final porosity,
        which is not easily specified since cylinders overlap and intersect different fractions of the domain
    :type nfibers: int, optional
    :param porosity: the targeted value for the porosity of the generated mat. The function uses an
        algorithm for predicted the number of required number of cylinder, and refines this over a certain number of
        fractional insertions (according to the 'iterations' input)
    :type porosity: float, optional
    :param phi: a value between 0 and 90 that controls the amount that the fibers lie *out of* the XY plane,
        with 0 meaning all fibers lie in the XY plane, and 90 meaning that cylinders are randomly oriented out of the
        plane by as much as +/- 90 degrees
    :type phi: float, optional
    :param theta: a value between 0 and 90 that controls the amount of rotation *in the* XY plane,
        with 0 meaning all fibers point in the X-direction, and 90 meaning they are randomly rotated about the
        Z axis by as much as +/- 90 degrees
    :type theta: float, optional
    :param length: the length of the cylinders to add.  If ``None`` (default) then the cylinders will extend beyond the
        domain in both directions so no ends will exist. If a scalar value is given it will be interpreted as the
        Euclidean distance between the two ends of the cylinder.  Note that one or both of the ends *may* still lie
        outside the domain, depending on the randomly chosen center point of the cylinder
    :type length: float, optional
    :param max_iter: the number of fractional fiber insertions used to target the requested porosity. By default a
        value of 3 is used (and this is typically effective in getting very close to the targeted porosity),
        but a greater number can be input to improve the achieved porosity
    :type max_iter: int, optional
    :return: random fibers domain
    :rtype: Workspace

    :Example:
    >>> import pumapy as puma
    >>> ws_fibers = puma.generate_random_fibers(shape=(200, 200, 200), radius=8, porosity=0.8, phi=90, theta=90, length=200)
    >>> puma.render_contour(ws_fibers, cutoff=(1, 1))
    >>> ws_fibers = puma.generate_random_fibers(shape=(200, 200, 200), radius=8, nfibers=100, phi=90, theta=90, length=200)
    >>> puma.render_contour(ws_fibers, cutoff=(1, 1))
    """

    # error checks
    if nfibers is None and porosity is None:
        raise Exception("'nfibers' and 'porosity' can't be both None")
    if max_iter < 3:
        raise Exception("Iterations must be >= 3")

    # run solver for provided number of fibers
    if nfibers is not None:
        img = _generate_fibers(shape=shape, radius=radius, nfibers=nfibers, phi=phi, theta=theta, length=length)

    else:  # porosity provided
        vol_total = float(np.prod(shape))

        length_estimate = vol_total ** (1 / 3) if length is None else length

        vol_fiber = length_estimate * np.pi * radius * radius
        n_pixels_to_add = -np.log(porosity) * vol_total

        n_fibers_added = 0

        subdif = 0.8 / np.sum(np.arange(1, max_iter) ** 2)
        fractions = [0.2]
        for i in range(1, max_iter):
            fractions.append(fractions[i - 1] + (max_iter - i) ** 2 * subdif)

        img = np.ones(shape, dtype=bool)
        for i, frac in enumerate(fractions):
            n_fibers_total = n_pixels_to_add / vol_fiber
            n_fibers = int(np.ceil(frac * n_fibers_total) - n_fibers_added)
            if n_fibers > 0:
                img = img & _generate_fibers(shape, radius, n_fibers, phi, theta, length)
            n_fibers_added += n_fibers

            void_vf = np.sum(img == 1)
            porosity = void_vf / (np.sum(img == 0) + void_vf)
            vol_added = -np.log(porosity) * vol_total
            vol_fiber = vol_added / n_fibers_added

            sys.stdout.write("\rGenerating fibers ... {:.1f}% ".format((i + 1) / len(fractions) * 100))

    img = np.where(img, np.uint16(0), np.uint16(1))
    img = Workspace.from_array(img.astype(np.uint16))
    print("\nGenerated random fibers domain with porosity: {}".format(compute_volume_fraction(img, (0, 0))))
    return img


def _generate_fibers(shape, radius, nfibers, phi=0, theta=90, length=None):
    """ Generates random fibers given nfibers"""

    # error checks
    shape = np.array(shape)
    if np.size(shape) == 1:
        shape = np.full((3, ), int(shape))
    elif np.size(shape) == 2:
        raise Exception("Shape can only be 3D")

    H = np.sqrt(np.sum(np.square(shape))).astype(int)
    if length is None:
        length = 2 * H
    R = min(int(length / 2), 2 * H)

    if (phi > 90) or (phi < 0):
        raise Exception('phi_max must be betwen 0 and 90')
    if (theta > 90) or (theta < 0):
        raise Exception('theta_max must be betwen 0 and 90')

    im = np.zeros(shape, dtype=bool)
    n = 0
    L = min(H, R)

    while n < nfibers:
        x = np.random.rand(3) * (shape + 2 * L)

        phi_rand = (np.pi / 2 - np.pi * np.random.rand()) * phi / 90
        theta_rand = (np.pi / 2 - np.pi * np.random.rand()) * theta / 90
        x0 = R * np.array([np.cos(phi_rand) * np.cos(theta_rand),
                           np.cos(phi_rand) * np.sin(theta_rand),
                           np.sin(phi_rand)])
        [x0, x1] = [x + x0, x - x0]

        x0 = np.around(x0).astype(int)
        x1 = np.around(x1).astype(int)
        L = np.amax(np.absolute([[x1[0] - x0[0]], [x1[1] - x0[1]], [x1[2] - x0[2]]])) + 1
        crds = [np.rint(np.linspace(x0[0], x1[0], L)).astype(int),
                np.rint(np.linspace(x0[1], x1[1], L)).astype(int),
                np.rint(np.linspace(x0[2], x1[2], L)).astype(int)]

        lower = ~np.any(np.vstack(crds).T < [L, L, L], axis=1)
        upper = ~np.any(np.vstack(crds).T >= shape + L, axis=1)
        valid = upper * lower
        if np.any(valid):
            im[crds[0][valid] - L, crds[1][valid] - L, crds[2][valid] - L] = 1
            n += 1

    im = np.array(im, dtype=bool)
    dt = edt(~im) < radius
    return ~dt



from pumapy.utilities.workspace import Workspace
from pumapy.utilities.isosurface import generate_isosurface
import pyvista as pv
import numpy as np


def render_volume(workspace, cutoff=None, solid_color=(1., 1., 1.), style='surface', origin=(0., 0., 0.),
                  window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=True,
                  show_axes=True, show_outline=True, cmap='gray', add_to_plot=None, notebook=False):
    """ Volume render using Pyvista Threshold filter

    :param workspace: domain
    :type workspace: Workspace or ndarray
    :param cutoff: specifying the values to render
    :type cutoff: tuple(int, int), optional
    :param solid_color: if set to None, the material is colored by the matrix's values. Otherwise, a solid color can be specified (e.g. for white (1., 1., 1.))
    :type solid_color: tuple(float, float, float), optional
    :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
    :type style: string, optional
    :param origin: origin of the data as
    :type origin: tuple(float, float, float), optional
    :param window_size: with the popup window size
    :type window_size: tuple(int, int), optional
    :param opacity: opacity of volume
    :type opacity: float, optional
    :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
    :type: background: tuple(float, float, float)
    :param show_grid: show the grid with the size of the sides
    :type show_grid: bool, optional
    :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
    :type plot_directly: bool, optional
    :param show_axes: show orientation axis in the bottom left corner
    :type show_axes: bool, optional
    :param show_outline: show the bounding box outline of the domain
    :type show_outline: bool, optional
    :param cmap: matplotlib colormap to use (overwritten by solid_color if specified)
    :type cmap: str, optional
    :param add_to_plot: pass an already existing plotter object to add on top of this plot
    :type add_to_plot: pyvista.Plotter, optional
    :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
    :type notebook: bool, optional
    :return: None is plot_directly is True, otherwise a plotter object
    :rtype: pyvista.Plotter object or None

    :Example
    >>> import pumapy as puma
    >>> ws_volume = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
    >>> puma.render_volume(ws_volume)
    """
    if cutoff is None:
        solid_color = None
    r = Renderer(add_to_plot, "threshold", workspace, cutoff, solid_color, style, origin, window_size, opacity,
                 background, show_grid, plot_directly, show_axes, show_outline, cmap, None, notebook)
    return r.render()


def render_contour(workspace, cutoff, solid_color=(1., 1., 1.), style='surface', origin=(0., 0., 0.),
                   window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True, plot_directly=True,
                   show_axes=True, show_outline=True, add_to_plot=None, notebook=False):
    """ Contour render using Pyvista Contour filter

    :param workspace: domain
    :type workspace: Workspace or ndarray
    :param cutoff: specifying the values at which the isosurface is created
    :type cutoff: tuple(int, int)
    :param solid_color: a solid color to color the surface (e.g. for white (1., 1., 1.))
    :type solid_color: tuple(float, float, float), optional
    :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
    :type style: string, optional
    :param origin: origin of the data as
    :type origin: tuple(float, float, float), optional
    :param window_size: with the popup window size
    :type window_size: tuple(int, int), optional
    :param opacity: opacity of contour
    :type opacity: float, optional
    :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
    :type: background: tuple(float, float, float)
    :param show_grid: show the grid with the size of the sides
    :type show_grid: bool, optional
    :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
    :type plot_directly: bool, optional
    :param show_axes: show orientation axis in the bottom left corner
    :type show_axes: bool, optional
    :param show_outline: show the bounding box outline of the domain
    :type show_outline: bool, optional
    :param add_to_plot: pass an already existing plotter object to add on top of this plot
    :type add_to_plot: pyvista.Plotter, optional
    :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
    :type notebook: bool, optional
    :return: None is plot_directly is True, otherwise a plotter object
    :rtype: pyvista.Plotter object or None

    :Example:
    >>> import pumapy as puma
    >>> ws_contour = puma.import_3Dtiff(puma.path_to_example_file("50_artfibers.tif"))
    >>> puma.render_contour(ws_contour, (128,255))
    """
    r = Renderer(add_to_plot, "contour", workspace, cutoff, solid_color, style, origin, window_size, opacity,
                 background, show_grid, plot_directly, show_axes, show_outline, None, None, notebook)
    return r.render()


def render_orientation(workspace, scale_factor=1., solid_color=(1., 1., 1.), style='surface', origin=(0., 0., 0.),
                       window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True,
                       plot_directly=True, show_axes=True, show_outline=True, add_to_plot=None, notebook=False):
    """ Orientation render using Pyvista Glyph filter

    :param workspace: domain
    :type workspace: Workspace or ndarray
    :param scale_factor: scale the arrows by a factor
    :type scale_factor: float
    :param solid_color: a solid color to color the surface (e.g. for white (1., 1., 1.))
    :type solid_color: tuple(float, float, float), optional
    :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
    :type style: string, optional
    :param origin: origin of the data as
    :type origin: tuple(float, float, float), optional
    :param window_size: with the popup window size
    :type window_size: tuple(int, int), optional
    :param opacity: opacity of arrows
    :type opacity: float, optional
    :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
    :type: background: tuple(float, float, float)
    :param show_grid: show the grid with the size of the sides
    :type show_grid: bool, optional
    :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
    :type plot_directly: bool, optional
    :param show_axes: show orientation axis in the bottom left corner
    :type show_axes: bool, optional
    :param show_outline: show the bounding box outline of the domain
    :type show_outline: bool, optional
    :param add_to_plot: pass an already existing plotter object to add on top of this plot
    :type add_to_plot: pyvista.Plotter, optional
    :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
    :type notebook: bool, optional
    :return: None is plot_directly is True, otherwise a plotter object
    :rtype: pyvista.Plotter object or None

    :Example:
    >>> import pumapy as puma
    >>> ws_orientation = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
    >>> puma.compute_orientation_st(ws_orientation, 0.7, 1.4, (90, 255))
    >>> puma.render_orientation(ws_orientation)
    """
    r = Renderer(add_to_plot, "glyph", workspace, None, solid_color, style, origin, window_size, opacity,
                 background, show_grid, plot_directly, show_axes, show_outline, None, scale_factor, notebook)
    return r.render()


def render_contour_multiphase(workspace, cutoffs, solid_colors=None, style='surface', origin=(0., 0., 0.),
                              window_size=(1920, 1200), opacity=1., background=(0.3, 0.3, 0.3), show_grid=True,
                              plot_directly=True, show_axes=True, show_outline=True, add_to_plot=None, notebook=False):
    """ Contour render for multi-phase materials using Pyvista

    :param workspace: domain
    :type workspace: Workspace or ndarray
    :param cutoffs: n cutoffs is the number of materials. specifies the low and high cutoff ranges
    :type cutoffs: tuple(tuple(int, int), tuple(int, int) ...)
    :param solid_colors: solid colors to color the different phases' surface e.g. for white ((1., 1., 1.), (0., 0., 0.), ...)
    :type solid_colors: tuple(tuple(float, float, float), tuple(float, float, float) ...), optional
    :param style: specifying the representation style ('surface', 'edges', 'wireframe', 'points')
    :type style: string, optional
    :param origin: origin of the data as
    :type origin: tuple(float, float, float), optional
    :param window_size: with the popup window size
    :type window_size: tuple(int, int), optional
    :param opacity: opacity of contour
    :type opacity: float, optional
    :param background: color of the background from (0., 0., 0.) (black) to (1., 1., 1.) (white)
    :type: background: tuple(float, float, float)
    :param show_grid: show the grid with the size of the sides
    :type show_grid: bool, optional
    :param plot_directly: whether to return a Plotter object (to make further changes to it) or show the plot directly
    :type plot_directly: bool, optional
    :param show_axes: show orientation axis in the bottom left corner
    :type show_axes: bool, optional
    :param show_outline: show the bounding box outline of the domain
    :type show_outline: bool, optional
    :param add_to_plot: pass an already existing plotter object to add on top of this plot
    :type add_to_plot: pyvista.Plotter, optional
    :param notebook: plotting interactively in a jupyter notebook (overwrites show_grid to False)
    :type notebook: bool, optional
    :return: None is plot_directly is True, otherwise a plotter object
    :rtype: pyvista.Plotter object or None

    >>> import pumapy as puma
    >>> ws_multiphase = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
    >>> puma.render_contour_multiphase(ws_multiphase, ((100, 150), (150, 255)))
    """

    if add_to_plot is None:
        p = pv.Plotter(notebook=notebook)
    else:
        p = add_to_plot
    if notebook:  # observed problems in rendering with grid in jupyter
        show_grid = False
    for i, cutoff in enumerate(cutoffs):
        if solid_colors is None or len(solid_colors) != len(cutoffs):
            solid_color = tuple(np.random.rand(3))
        else:
            solid_color = solid_colors[i]
        r = Renderer(p, "contour", workspace, cutoff, solid_color, style, origin, window_size, opacity,
                     background, show_grid, False, show_axes, show_outline, None, None, None)
        p = r.render()
    if plot_directly:
        p.show()
    else:
        return p


class Renderer:
    def __init__(self, existing_plot, filter_type, workspace, cutoff, solid_color, style, origin, window_size, opacity,
                 background, show_grid, plot_directly, show_axes, show_outline, cmap, scale_factor, notebook):
        self.filter_type = filter_type
        self.cutoff = cutoff
        self.solid_color = solid_color
        self.style = style
        self.origin = origin
        self.window_size = window_size
        self.opacity = opacity
        self.background = background
        self.show_grid = show_grid
        self.plot_directly = plot_directly
        self.show_axes = show_axes
        self.show_outline = show_outline
        self.cmap = cmap
        self.scale_factor = scale_factor

        if isinstance(workspace, Workspace):
            if self.filter_type == "glyph":
                self.array = workspace.orientation
            else:
                self.array = workspace.matrix
        elif isinstance(workspace, np.ndarray):
            if self.filter_type == "glyph":
                if not (workspace.ndim == 4 and workspace.shape[3] == 3):
                    raise Exception("For rendering orientation, need a numpy.ndarray of shape (X, Y, Z, 3)")
            self.array = workspace
        else:
            raise Exception("Need to input either pumapy.Workspace or a numpy.ndarray")

        if self.filter_type != "contour":
            self.grid = pv.UniformGrid()
            self.grid.origin = self.origin

        self.filter = None
        if existing_plot is None:
            self.p = pv.Plotter(notebook=notebook)
        else:
            self.p = existing_plot

        if notebook:  # observed problems in rendering with grid in jupyter
            self.show_grid = False

    def render(self):
        self.build_plotter()

        if self.plot_directly:
            self.p.show()
        else:
            return self.p

    def build_plotter(self):
        if self.style == 'edges':
            show_edges = True
            self.style = 'surface'
        else:
            show_edges = False

        if self.filter_type == "threshold":
            self.grid.dimensions = np.array(self.array.shape) + 1
            self.grid.cell_arrays["values"] = self.array.flatten(order="F")
            self.filter = self.grid.threshold(self.cutoff)

        elif self.filter_type == "contour":
            tri_mesh = generate_isosurface(self.array, self.cutoff)
            tri_mesh.verts += self.origin
            tri_mesh.create_mesh()
            self.filter = tri_mesh.mesh

        elif self.filter_type == "glyph":
            self.grid.dimensions = np.array(self.array.shape[:3]) + 1
            tmp = np.zeros((self.array[:, :, :, 0].size, 3), dtype=float)
            for i in range(3):
                tmp[:, i] = self.array[:, :, :, i].ravel(order='F')
            self.grid.cell_arrays["scalars"] = np.linalg.norm(self.array, axis=3).ravel(order='F')
            self.grid.cell_arrays["vectors"] = tmp
            self.filter = self.grid.glyph(orient="vectors", scale="scalars", factor=self.scale_factor, geom=pv.Arrow())

        self.p.add_mesh(self.filter, color=self.solid_color, show_edges=show_edges, style=self.style,
                        cmap=self.cmap, show_scalar_bar=False, opacity=self.opacity)

        if self.cmap is not None and self.solid_color is None:
            self.p.add_scalar_bar(title='', height=0.5, vertical=True, position_x=0.05, position_y=0.3)

        self.p.background_color = self.background
        self.p.window_size = self.window_size
        if self.show_outline:
            self.p.add_bounding_box(color="k")
        if self.show_axes:
            self.p.show_axes()
        if self.show_grid:
            self.p.show_grid()



from pumapy.physicsmodels.isotropic_conductivity import IsotropicConductivity
from pumapy.physicsmodels.mpfa_conductivity import AnisotropicConductivity
from pumapy.utilities.property_maps import IsotropicConductivityMap, AnisotropicConductivityMap


def compute_thermal_conductivity(workspace, cond_map, direction, side_bc='s', prescribed_bc=None, tolerance=1e-4,
                                 maxiter=10000, solver_type='bicgstab', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the thermal conductivity

    :param workspace: domain
    :type workspace: Workspace
    :param cond_map: local constituents themal conductivities
    :type cond_map: IsotropicConductivityMap or AnisotropicConductivityMap
    :param direction: direction for solve ('x','y', or 'z')
    :type direction: string
    :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p') or dirichlet ('d')
    :type side_bc: string, optional
    :param prescribed_bc: 3D array holding dirichlet BC.
    :type prescribed_bc: ConductivityBC, optional
    :param tolerance: tolerance for iterative solver
    :type tolerance: float, optional
    :param maxiter: maximum Iterations for solver
    :type maxiter: int, optional
    :param solver_type: solver type, options: 'bicgstab', 'cg', 'gmres', 'direct'
    :type solver_type: string, optional
    :param display_iter: display iterations and residual
    :type display_iter: bool, optional
    :param print_matrices: corresponding to b, E, A, T, q decimal places. If 0, they are not printed
    :type print_matrices: tuple(5 bools), optional
    :return: thermal conductivity, temperature field, flux
    :rtype: tuple(tuple(float, float, float), ndarray, ndarray)

    :Example:
    >>> import pumapy as puma
    >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
    >>> ws_fiberform.matrix = ws_fiberform.matrix[50:150, 50:150, 50:150]
    >>> cond_map = puma.IsotropicConductivityMap()
    >>> cond_map.add_material((0, 89), 0.0257)
    >>> cond_map.add_material((90, 255), 12)
    >>> k_eff_x, T_x, q_x = puma.compute_thermal_conductivity(ws_fiberform, cond_map, 'x', 's', tolerance=1e-2, solver_type='cg')
    >>> print("Effective thermal conductivity tensor:")
    >>> print(k_eff_x)
    """
    if isinstance(cond_map, IsotropicConductivityMap):
        solver = IsotropicConductivity(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                                       solver_type, display_iter)
    elif isinstance(cond_map, AnisotropicConductivityMap):
        solver = AnisotropicConductivity(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                                         solver_type, display_iter, print_matrices)
    else:
        raise Exception("cond_map has to be an IsotropicConductivityMap or AnisotropicConductivityMap")

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.keff, solver.T, solver.q


def compute_electrical_conductivity(workspace, cond_map, direction, side_bc='p', prescribed_bc=None, tolerance=1e-4,
                                    maxiter=10000, solver_type='bicgstab', display_iter=True, print_matrices=(0, 0, 0, 0, 0)):
    """ Compute the electrical conductivity

    :param workspace: domain
    :type workspace: Workspace
    :param cond_map: local constituents electrical conductivities
    :type cond_map: IsotropicConductivityMap or AnisotropicConductivityMap
    :param direction: direction for solve ('x','y', or 'z')
    :type direction: string
    :param side_bc: side boundary conditions can be symmetric ('s'), periodic ('p') or dirichlet ('d')
    :type side_bc: string, optional
    :param prescribed_bc: 3D array holding dirichlet BC
    :type prescribed_bc: ConductivityBC, optional
    :param tolerance: tolerance for iterative solver
    :type tolerance: float, optional
    :param maxiter: maximum Iterations for solver
    :type maxiter: int, optional
    :param solver_type: solver type, options: 'bicgstab', 'cg', 'gmres', 'direct'
    :type solver_type: string, optional
    :param display_iter: display iterations and residual
    :type display_iter: bool, optional
    :param print_matrices: corresponding to E, A, b, T, q decimal places. If 0, they are not printed
    :type print_matrices: tuple(5 bools), optional
    :return: electrical conductivity, potential field, flux
    :rtype: tuple(tuple(float, float, float), ndarray, ndarray)

    :Example:
    >>> import pumapy as puma
    >>> ws_fiberform = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"), 1.3e-6)
    >>> ws_fiberform.matrix = ws_fiberform.matrix[50:150, 50:150, 50:150]
    >>> cond_map = puma.IsotropicConductivityMap()
    >>> cond_map.add_material((0, 89), 0.0257)
    >>> cond_map.add_material((90, 255), 12)
    >>> k_eff_x, P_x, q_x = puma.compute_electrical_conductivity(ws_fiberform, cond_map, 'x', 's', tolerance=1e-2, solver_type='cg')
    >>> print("Effective electrical conductivity tensor:")
    >>> print(k_eff_x)
    """
    return compute_thermal_conductivity(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter,
                                        solver_type, display_iter, print_matrices)



import numpy as np
from skimage import measure
from pumapy.utilities.generic_checks import check_ws_cutoff


def identify_porespace(workspace, solid_cutoff):
    """ Identify the porespace

        :param workspace: domain
        :type workspace: Workspace
        :param solid_cutoff: specify the solid range to discard from pores identification
        :type solid_cutoff: tuple(int, int)
        :return: porespace marked as: 0 solid, 1 largest pore (likely open porosity), >1 other pores
        :rtype: ndarray

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_random_spheres((200, 200, 200), diameter=20, porosity=0.8, allow_intersect=True)
        >>> ws.binarize_range((0, 128))  # invert material, i.e. consider spheres as pores
        >>> pores = puma.identify_porespace(ws, (1, 1))
        >>> puma.render_volume(pores, (1, pores.max()), solid_color=None, cmap='jet')
    """

    # error check
    check_ws_cutoff(workspace, solid_cutoff)
    ws = workspace.copy()

    ws.binarize_range(solid_cutoff)

    pore_labels = measure.label(ws.matrix, background=1)

    unique_pore_ids, unique_id_counts = np.unique(pore_labels[pore_labels != 0], return_counts=True)

    sorted_unique_pore_ids = unique_pore_ids[np.argsort(unique_id_counts)[::-1]]

    keyarray = np.arange(np.max(pore_labels) + 1)
    keyarray[sorted_unique_pore_ids] = unique_pore_ids

    return keyarray[pore_labels]


def fill_closed_pores(workspace, solid_cutoff, fill_value, return_pores=False):
    """ Identify the porespace and fill closed porosity

        :param workspace: domain
        :type workspace: Workspace
        :param solid_cutoff: specify the solid range to discard from pores identification
        :type solid_cutoff: tuple(int, int)
        :param fill_value: value to fill closed porosity with
        :type fill_value: int
        :param return_pores: specifies whether to return identified pores
        :type return_pores: bool, optional
        :return: filled workspace with IDs set as: 0 for void, 1 for solid, fill_value for 
            added filler material. In addition, if return_pores==True, then it also returns 
            the porespace marked as: 0 solid, 1 largest pore (likely open porosity), >1 other pores
        :rtype: Workspace

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_random_spheres((200, 200, 200), diameter=20, porosity=0.8, allow_intersect=True)
        >>> ws.binarize_range((0, 128))  # invert material, i.e. consider spheres as pores
        >>> filled_ws, pores = puma.fill_closed_pores(ws, solid_cutoff=(1, 1), fill_value=2, return_pores=True)
        >>> puma.render_volume(pores, (1, pores.max()), solid_color=None, cmap='jet')
        >>> puma.render_volume(filled_ws)
    """

    pores = identify_porespace(workspace, solid_cutoff)

    if isinstance(pores, bool):
        return False

    ws = workspace.copy()
    ws.binarize_range(solid_cutoff)

    ws[np.where(pores > 1)] = np.uint16(fill_value)

    if return_pores:
        return ws, pores
    else:
        return ws



import unittest
import numpy as np
import pumapy as puma


class TestRadiation(unittest.TestCase):

    def test_nophase(self):
        ws = puma.Workspace.from_shape_value((10, 10, 10), 1)
        try:
            puma.compute_radiation(ws, (1, 1), 1, 45)
        except:
            print("Correctly caught exception.")
            return
        self.assertTrue(False)

    def test_basic(self):
        ws = puma.Workspace.from_shape_value((10, 10, 10), 1)
        ws[4, 4, 4] = 0
        self.assertTrue(puma.compute_radiation(ws, (1, 1), 1, 45) is not False)

    def test_artfib(self):
        ws = puma.import_vti("testdata/artifib.vtk")
        ws.set_material_id((0, 128), 0)
        ws.set_material_id((129, 255), 1)

        np.random.seed(0)
        beta, beta_std, _ = puma.compute_radiation(ws, (1, 1), 100, 15, boundary_behavior=0)
        print(beta)
        np.testing.assert_almost_equal(beta, [0.18048920415267472, 0.17922697533941973, 0.1241739185326434], decimal=4)


if __name__ == '__main__':
    unittest.main()



from TexGen.Core import *


def export_weave_vtu(filename, weave, domain, max_dim_nvox, round_vox_up=True, export_orientation=True):
    """ Exporting weave to vtu, to be read by pumapy

    :param filename: filepath and name
    :type filename: string
    :param weave: weave object, as defined in TexGen
    :type weave: CTextile or child class of CTextile
    :param domain: domain size object, as defined in TexGen
    :type domain: CDomainPlanes
    :param max_dim_nvox: number of voxels to add in the largest domain dimension
    :type max_dim_nvox: int
    :param round_vox_up: for the shorter dimensions, round number of voxels up (for +/-1 vox)
    :type round_vox_up: bool
    :param export_orientation: specify whether to export orientation
    :type export_orientation: bool
    :return: filename of weave exported (input filename + dimensions)
    :rtype: string
    """

    if not isinstance(domain, CDomainPlanes):
        raise Exception("Domain needs to be of CDomainPlanes type.")
    if not isinstance(filename, str):
        raise Exception("Filename has to be a string.")
    if not path.exists(path.split(filename)[0]):
        raise Exception("Directory " + path.split(filename)[0] + " not found.")

    min_bounds = XYZ()
    max_bounds = XYZ()
    domain.GetBoxLimits(min_bounds, max_bounds)

    weave.AssignDomain(CDomainPlanes(min_bounds, max_bounds))

    lengths = np.array([max_bounds.x - min_bounds.x, max_bounds.y - min_bounds.y, max_bounds.z - min_bounds.z])
    max_len = np.max(lengths)

    mask = np.zeros(3, dtype=bool)
    mask[lengths == max_len] = True

    voxel_length = max_len / float(max_dim_nvox)

    nvox = np.zeros(3, dtype=int)
    nvox[mask] = max_dim_nvox
    nvox[~mask] = (lengths[~mask] / voxel_length).astype(int)  # truncates

    rem = np.zeros(3, dtype=float)
    rem[~mask] = lengths[~mask] - voxel_length * nvox[~mask]

    if round_vox_up:
        rem[~mask] = voxel_length - rem[~mask]
        max_bounds = XYZ(max_bounds.x + rem[0],
                         max_bounds.y + rem[1],
                         max_bounds.z + rem[2])
        nvox[~mask] += 1
    else:
        max_bounds = XYZ(max_bounds.x - rem[0], max_bounds.y - rem[1], max_bounds.z - rem[2])
    weave.AssignDomain(CDomainPlanes(min_bounds, max_bounds))

    mesh = CRectangularVoxelMesh()
    print("Exporting " + filename + ".vtu ... ", end='')
    filename += "_" + str(nvox[0]) + "_" + str(nvox[1]) + "_" + str(nvox[2])
    mesh.SaveVoxelMesh(weave, filename, int(nvox[0]), int(nvox[1]), int(nvox[2]), False, export_orientation,
                       MATERIAL_CONTINUUM, 0, VTU_EXPORT)
    print("Done")
    return filename



import unittest
import numpy as np
import pumapy as puma
from skimage import io


class TestIsosurface(unittest.TestCase):

    def test_fiberform(self):
        ws = puma.Workspace.from_array(io.imread(puma.path_to_example_file("200_fiberform.tif")))
        mesh = puma.generate_isosurface(ws, (90, 255), True, False)

        np.testing.assert_equal(len(mesh.values), 665453)
        np.testing.assert_equal(mesh.faces.shape, (1330866, 3))


if __name__ == '__main__':
    unittest.main()



from pumapy.physicsmodels.anisotropic_conductivity_utils import (pad_domain, add_nondiag, divP,
                                                                 fill_flux, flatten_Kmat_find_unstable_iv)
from pumapy.physicsmodels.mpxa_matrices import fill_Ampfa, fill_Bmpfa, fill_Cmpfa, fill_Dmpfa, create_mpfa_indices
from pumapy.physicsmodels.conductivity_parent import Conductivity
from pumapy.utilities.timer import Timer
from scipy.sparse import csr_matrix, diags
import numpy as np
import sys


class AnisotropicConductivity(Conductivity):

    def __init__(self, workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter, solver_type,
                 display_iter, print_matrices):
        super().__init__(workspace, cond_map, direction, side_bc, prescribed_bc, tolerance, maxiter, solver_type, display_iter)
        self.print_matrices = print_matrices
        self.mat_cond = dict()
        self.need_to_orient = False  # changes if conductivities (k_axial, k_radial) detected
        self.orient_pad = None

    def compute(self):
        t = Timer()
        self.initialize()
        self.assemble_bvector()
        self.assemble_Amatrix()
        print("Time to assemble matrices: ", t.elapsed()); t.reset()
        super().solve()
        print("Time to solve: ", t.elapsed())
        self.compute_effective_coefficient()
        self.solve_time = t.elapsed()

    def initialize(self):
        print("Initializing and padding domains ... ", flush=True, end='')

        # Rotating domain to avoid cases and padding
        shape = [self.len_x + 2, self.len_y + 2, self.len_z + 2]
        reorder = [0, 1, 2]
        reorder_nondiagcond = [3, 4, 5]
        if self.direction == 'y':
            shape = [self.len_y + 2, self.len_x + 2, self.len_z + 2]
            reorder = [1, 0, 2]
            reorder_nondiagcond = [3, 5, 4]
        elif self.direction == 'z':
            shape = [self.len_z + 2, self.len_y + 2, self.len_x + 2]
            reorder = [2, 1, 0]
            reorder_nondiagcond = [5, 4, 3]

        self.ws_pad = np.zeros(shape, dtype=np.uint16)
        self.ws_pad[1:-1, 1:-1, 1:-1] = self.ws.matrix.transpose(reorder)

        if self.need_to_orient:
            self.orient_pad = np.zeros(shape + [3], dtype=float)
            self.orient_pad[1:-1, 1:-1, 1:-1, :] = self.ws.orientation.transpose(reorder + [3])[:, :, :, reorder]

        for key in self.mat_cond.keys():
            tmp_cond = list(self.mat_cond[key])
            if len(tmp_cond) == 6:
                self.mat_cond[key] = tuple([x for _, x in sorted(zip(reorder + reorder_nondiagcond, tmp_cond))])

        self.len_x, self.len_y, self.len_z = shape
        self.len_xy = self.len_x * self.len_y
        self.len_xyz = self.len_x * self.len_y * self.len_z

        # Padding domain, imposing symmetric or periodic BC on faces
        pad_domain(self.ws_pad, self.orient_pad, self.need_to_orient, self.len_x, self.len_y, self.len_z, self.side_bc)

        # Segmenting padded domain
        for i in range(self.cond_map.get_size()):
            low, high, _ = self.cond_map.get_material(i)
            self.ws_pad[(self.ws_pad >= low) & (self.ws_pad <= high)] = i

        # Placing True on dirichlet boundaries to skip them
        self.dir_vox = np.zeros(shape, dtype=np.uint8)
        self.dir_vox[[1, -2], 1:-1, 1:-1] = 1
        if self.prescribed_bc is not None:
            self.dir_vox[1:-1, 1:-1, 1:-1][self.prescribed_bc.dirichlet != np.Inf] = 1
        print("Done")

        # Initialize initial guess for iterative solver
        if self.solver_type != 'direct' and self.solver_type != 'spsolve':
            self.initial_guess = np.zeros((self.len_x, self.len_y, self.len_z), dtype=float)
            for i in range(self.len_x - 1):
                self.initial_guess[i] = i / (self.len_x - 2.)
            self.initial_guess = self.initial_guess.flatten('F')

    def assemble_bvector(self):
        print("Assembling b vector ... ", flush=True, end='')

        I, V = ([] for _ in range(2))

        if self.prescribed_bc is not None:
            for i in range(1, self.len_x - 1):
                for j in range(1, self.len_y - 1):
                    for k in range(1, self.len_z - 1):
                        if self.prescribed_bc[i - 1, j - 1, k - 1] != np.Inf:
                            I.append(self.len_x * (self.len_y * k + j) + i)
                            V.append(self.prescribed_bc[i - 1, j - 1, k - 1])
        else:
            # Setting unit temperature
            i = self.len_x - 2
            for j in range(1, self.len_y - 1):
                for k in range(1, self.len_z - 1):
                    I.append(self.len_x * (self.len_y * k + j) + i)
                    V.append(1.)

        # Setting linear temperature on the boundaries if Dirichlet
        if self.side_bc == 'd':
            x = np.linspace(0, 1, self.len_x - 2)
            for j in [1, self.len_y - 2]:
                for i in range(1, self.len_x - 1):
                    for k in range(1, self.len_z - 1):
                        I.append(self.len_x * (self.len_y * k + j) + i)
                        V.append(x[i - 1])
            for k in [1, self.len_z - 2]:
                for i in range(2, self.len_x - 2):
                    for j in range(2, self.len_y - 2):
                        I.append(self.len_x * (self.len_y * k + j) + i)
                        V.append(x[i - 1])

        self.bvec = csr_matrix((V, (I, np.zeros(len(I)))), shape=(self.len_xyz, 1))

        if self.print_matrices[0]:
            self._print_b(self.print_matrices[0])
        print("Done")


    def assemble_Amatrix(self):
        print("Initializing large data structures ... ", flush=True, end='')
        I, J = np.zeros((2, 27 * self.len_xyz), dtype=np.uint32)
        V = np.zeros(27 * self.len_xyz, dtype=float)
        counter = 0
        I_dirvox = []
        self.__initialize_MPFA()
        j_indices = np.zeros((27 * (self.len_y - 2) * (self.len_z - 2)), dtype=np.uint32)
        values = np.zeros((27 * (self.len_y - 2) * (self.len_z - 2)), dtype=float)
        print("Done")

        # Iterating through interior
        for i in range(1, self.len_x - 1):
            self.__compute_Kmat(2, i + 1)  # Computing third layer of Kmat
            self.__compute_transmissibility(1, i)  # Computing second layer of E

            # Creating j indices and divergence values for slice
            j_indices.fill(0)
            values.fill(np.NaN)

            # compute flux divergence: if a row of A is detected to be all zeros, then the voxel is flagged as dirichlet
            divP(i, self.len_x, self.len_y, self.len_z, self.dir_vox, j_indices, values, self.Emat)

            # Creating i indices for slice
            i_indices, i_dirvox = self.__creating_indices(i)
            if i_indices.size > 0:
                I[counter:counter + i_indices.size] = i_indices
            I_dirvox.extend(i_dirvox)

            if j_indices[j_indices != 0].size > 0:
                J[counter:counter + i_indices.size] = j_indices[~np.isnan(values)]
                V[counter:counter + i_indices.size] = values[~np.isnan(values)]
                counter += i_indices.size

            # Passing second layer to first
            self.Emat[0] = self.Emat[1]
            self.Kmat[:2] = self.Kmat[1:]
            sys.stdout.write("\rAssembling A matrix ... {:.1f}% ".format(i / (self.len_x - 2) * 100))

        # Clear unnecessary variables before creating A
        del self.Emat, self.kf, self.Kmat, self.unstable
        del self.dir_vox, i_indices, j_indices, values, i_dirvox, i

        # Adding all dirichlet voxels
        I[counter:counter + len(I_dirvox)] = I_dirvox
        J[counter:counter + len(I_dirvox)] = I_dirvox
        V[counter:counter + len(I_dirvox)] = 1
        counter += len(I_dirvox)
        del I_dirvox

        # Add diagonal 1s for exterior voxels
        diag_1s = np.ones_like(self.ws_pad, dtype=int)
        diag_1s[1:-1, 1:-1, 1:-1] = 0
        ind = np.array(np.where(diag_1s > 0))
        diag_1s = self.len_x * (self.len_y * ind[2] + ind[1]) + ind[0]
        diag_1s = diag_1s.astype(np.uint32)
        del ind
        I[counter:counter + diag_1s.size] = diag_1s
        J[counter:counter + diag_1s.size] = diag_1s
        V[counter:counter + diag_1s.size] = 1
        counter += diag_1s.size

        # Add non-diagonal 1s for exterior voxels
        if self.side_bc is not "d":
            I[counter:counter + diag_1s.size] = diag_1s
            add_nondiag(diag_1s, self.len_x, self.len_y, self.len_z, self.side_bc)
            J[counter:counter + diag_1s.size] = diag_1s
            V[counter:counter + diag_1s.size] = -1
            counter += diag_1s.size
        del diag_1s, counter

        # Assemble sparse A matrix
        self.Amat = csr_matrix((V, (I, J)), shape=(self.len_xyz, self.len_xyz))

        # Simple preconditioner
        diag = self.Amat.diagonal()
        if np.any(diag == 0):
            self.M = None  # identity matrix if singularity has happened in MPFA
        else:
            self.M = diags(1. / self.Amat.diagonal(), 0).tocsr()

        if self.print_matrices[2]:
            self._print_A(self.print_matrices[2])
        print("Done")

    def compute_effective_coefficient(self):
        # reshaping solution
        self.T = self.x.reshape([self.len_x, self.len_y, self.len_z], order='F')
        del self.x

        # Mirroring boundaries for flux computation
        self.T[0] = self.T[1]
        self.T[-1] = self.T[-2]
        if self.side_bc == "d":
            self.T[:, 0] = self.T[:, 1]
            self.T[:, -1] = self.T[:, -2]
            self.T[:, :, 0] = self.T[:, :, 1]
            self.T[:, :, -1] = self.T[:, :, -2]

        self.__compute_fluxes()

        # Accumulating and volume averaging stresses
        fluxes = [np.sum(self.q[:, :, :, i]) / ((self.len_x - 2) * (self.len_y - 2) * (self.len_z - 2))
                  for i in range(3)]
        self.keff = [-fluxes[i] * (self.len_x - 2) * self.ws.voxel_length for i in range(3)]

        # Rotating output back
        if self.direction == 'y':
            self.T = self.T.transpose(1, 0, 2)
            self.q = self.q.transpose(1, 0, 2, 3)[:, :, :, [1, 0, 2]]
            self.keff = [self.keff[1], self.keff[0], self.keff[2]]
        elif self.direction == 'z':
            self.T = self.T.transpose(2, 1, 0)
            self.q = self.q.transpose(2, 1, 0, 3)[:, :, :, [2, 1, 0]]
            self.keff = [self.keff[2], self.keff[1], self.keff[0]]

        d = {'x': 'first', 'y': 'second', 'z': 'third'}
        print(f'\nEffective conductivity tensor ({d[self.direction]} row): \n{self.keff}')

    def __compute_Kmat(self, i, i_cv):
        # reset layer of Cmat
        self.Kmat[i].fill(0)

        for key, value in self.mat_cond.items():
            mask = self.ws_pad[i_cv] == key
            if len(value) == 6:  # passing input conductivity
                self.Kmat[i, mask] = value
            else:  # tensor rotation of input conductivity with orientation
                phi = np.arctan2(self.orient_pad[i_cv, mask, 1], self.orient_pad[i_cv, mask, 0])
                theta = np.arcsin(self.orient_pad[i_cv, mask, 2])

                size = np.sum(mask)
                Rz_kinit = np.zeros((size, 3, 3), dtype=float)
                Ry_krot = np.zeros((size, 3, 3), dtype=float)

                Rz_kinit[:, 0, 0] = np.cos(phi)
                Rz_kinit[:, 1, 1] = Rz_kinit[:, 0, 0]
                Rz_kinit[:, 1, 0] = np.sin(phi)
                Rz_kinit[:, 0, 1] = -Rz_kinit[:, 1, 0]
                Rz_kinit[:, 2, 2] = 1
                Ry_krot[:, 1, 1] = 1
                Ry_krot[:, 0, 0] = np.cos(theta)
                Ry_krot[:, 2, 2] = Ry_krot[:, 0, 0]
                Ry_krot[:, 0, 2] = np.sin(theta)
                Ry_krot[:, 2, 0] = -Ry_krot[:, 0, 2]

                R = Rz_kinit @ Ry_krot

                Rz_kinit.fill(0)
                Rz_kinit[:, [0, 1, 2], [0, 1, 2]] = [value[0], value[1], value[1]]

                Ry_krot = R @ Rz_kinit @ np.linalg.inv(R)
                self.Kmat[i, mask] = Ry_krot[:, [0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]]

    def __compute_transmissibility(self, i, i_cv):
        # reset layers
        self.Emat[i].fill(0)
        self.unstable[i].fill(False)
        self.kf.fill(0)

        # if any of the surrounding CV have a zero diagonal cond, then set IV as unstable to skip computation
        flatten_Kmat_find_unstable_iv(self.len_y, self.len_z, self.Kmat[i:i + 2], self.kf, self.unstable[i])

        # creating C
        self.mpfa12x12.fill(0)
        self.mpfa12x12[:, :, self.Cind[0], self.Cind[1]] = fill_Cmpfa(self.kf)

        # Computing transmissibility matrix as: A @ (Cinv @ D) + B
        if not np.all(self.unstable[i]):

            # creating D
            self.Emat[i, :, :, self.Dind[0], self.Dind[1]] = fill_Dmpfa(self.kf)
            self.Emat[i, self.unstable[i]] = 0

            # computing:  Cinv
            self.mpfa12x12[~self.unstable[i]] = np.linalg.inv(self.mpfa12x12[~self.unstable[i]])

            # computing:  Cinv @ D
            self.Emat[i, ~self.unstable[i]] = self.mpfa12x12[~self.unstable[i]] @ self.Emat[i, ~self.unstable[i]]

            # creating A
            self.mpfa12x12.fill(0)
            self.mpfa12x12[:, :, self.Aind[0], self.Aind[1]] = fill_Ampfa(self.kf)

            # computing:  A @ (Cinv @ D)
            self.Emat[i, ~self.unstable[i]] = self.mpfa12x12[~self.unstable[i]] @ self.Emat[i, ~self.unstable[i]]

            # creating and adding B:  A @ (Cinv @ D) + B
            self.Emat[i, ~self.unstable[i]] += fill_Bmpfa(self.kf, self.zeros)[~self.unstable[i]]

        if self.print_matrices[1]:
            self._print_E(i, i_cv, self.print_matrices[1])

    def __initialize_MPFA(self):
        # Initialize matrix slice of conductivities
        self.Kmat = np.zeros((3, self.len_y, self.len_z, 6), dtype=float)  # per CV: kxx, kyy, kzz, kxy, kxz, kyz
        self.__compute_Kmat(0, 0)  # Computing first layer of Kmat
        self.__compute_Kmat(1, 1)  # Computing second layer of Kmat

        # Initialize MPFA variables
        self.kf = np.zeros((48, self.len_y - 1, self.len_z - 1), dtype=float)  # per IV
        self.Emat = np.zeros((2, self.len_y - 1, self.len_z - 1, 12, 8), dtype=float)
        self.unstable = np.zeros((2, self.len_y - 1, self.len_z - 1), dtype=bool)
        self.mpfa12x12 = np.zeros((self.len_y - 1, self.len_z - 1, 12, 12), dtype=float)  # A, C
        self.zeros = np.zeros(self.kf[0].shape)
        self.Aind, self.Cind, self.Dind = create_mpfa_indices()
        self.__compute_transmissibility(0, 0)  # Computing first layer of E

    def __creating_indices(self, i):
        # Finding all indices for slice
        i_indices = np.ones_like(self.ws_pad[i], dtype=np.uint32)
        i_indices[[0, -1], :] = 0
        i_indices[:, [0, -1]] = 0
        i_indices = np.where(i_indices > 0)
        i_indices = self.len_x * (self.len_y * i_indices[1] + i_indices[0]) + np.full(i_indices[0].size, i)

        # Removing dirichlet voxels
        i_dirvox = np.where(self.dir_vox[i])
        i_dirvox = self.len_x * (self.len_y * i_dirvox[1] + i_dirvox[0]) + np.full(i_dirvox[0].size, i)
        i_indices = i_indices[~np.in1d(i_indices, i_dirvox)]

        # Duplicating the voxel indices where divergence happens
        i_indices = np.repeat(i_indices, 27)
        return i_indices, i_dirvox  # returning dirichlet voxel indices

    def __compute_fluxes(self):
        # Initialize required data structures
        self.q = np.zeros((self.len_x - 2, self.len_y - 2, self.len_z - 2, 3), dtype=float)
        self.__initialize_MPFA()
        T_sw, T_se, T_nw, T_ne, T_tsw, T_tse, T_tnw, T_tne = np.zeros((8, self.len_y - 2, self.len_z - 2, 8))
        E_sw, E_se, E_nw, E_ne, E_tsw, E_tse, E_tnw, E_tne = np.zeros((8, self.len_y - 2, self.len_z - 2, 12, 8))

        # Iterating through interior
        for i in range(1, self.len_x - 1):
            self.__compute_Kmat(2, i + 1)  # Computing third layer of Kmat
            self.__compute_transmissibility(1, i)  # Computing second layer of E

            # filling eight IVs
            fill_flux(i, self.len_x, self.len_y, self.len_z, self.T[i - 1:i + 2], self.Emat,
                               E_sw, E_se, E_nw, E_ne, E_tsw, E_tse, E_tnw, E_tne,
                               T_sw, T_se, T_nw, T_ne, T_tsw, T_tse, T_tnw, T_tne)

            # Computing fluxes by computing E x T
            q_sw = np.squeeze(E_sw @ T_sw[:, :, :, np.newaxis])
            q_se = np.squeeze(E_se @ T_se[:, :, :, np.newaxis])
            q_nw = np.squeeze(E_nw @ T_nw[:, :, :, np.newaxis])
            q_ne = np.squeeze(E_ne @ T_ne[:, :, :, np.newaxis])
            q_tsw = np.squeeze(E_tsw @ T_tsw[:, :, :, np.newaxis])
            q_tse = np.squeeze(E_tse @ T_tse[:, :, :, np.newaxis])
            q_tnw = np.squeeze(E_tnw @ T_tnw[:, :, :, np.newaxis])
            q_tne = np.squeeze(E_tne @ T_tne[:, :, :, np.newaxis])

            # Summing fluxes CV-wise
            self.q[i - 1, :, :, 0] = (q_se[:, :, 3] + q_ne[:, :, 2] + q_tse[:, :, 1] + q_tne[:, :, 0] +
                                      q_sw[:, :, 3] + q_nw[:, :, 2] + q_tsw[:, :, 1] + q_tnw[:, :, 0]) / 4
            self.q[i - 1, :, :, 1] = (q_nw[:, :, 7] + q_ne[:, :, 6] + q_tnw[:, :, 5] + q_tne[:, :, 4] +
                                      q_sw[:, :, 7] + q_se[:, :, 6] + q_tsw[:, :, 5] + q_tse[:, :, 4]) / 4
            self.q[i - 1, :, :, 2] = (q_tsw[:, :, 11] + q_tse[:, :, 10] + q_tnw[:, :, 9] + q_tne[:, :, 8] +
                                      q_sw[:, :, 11] + q_se[:, :, 10] + q_nw[:, :, 9] + q_ne[:, :, 8]) / 4

            # Passing second layer to first
            self.Emat[0] = self.Emat[1]
            self.Kmat[0:2] = self.Kmat[1:3]
            sys.stdout.write("\rComputing fluxes ... {:.1f}% ".format(i / (self.len_x - 2) * 100))
        del self.Emat, self.kf, self.Kmat, self.unstable

        # Extract only interior temperature, ignoring exterior used as bc
        self.T = self.T[1:-1, 1:-1, 1:-1]
        self.q /= self.ws.voxel_length

        if self.print_matrices[3]:
            print_T(self.T, self.print_matrices[3])
        if self.print_matrices[4]:
            print_flux(self.q, self.print_matrices[4])
        print("Done")

    def error_check(self):
        if Conductivity.error_check(self):
            return False

        # cond_map checks
        ws_tmp_tocheck = self.ws.matrix.copy()
        for i in range(self.cond_map.get_size()):
            low, high, k = self.cond_map.get_material(i)
            self.mat_cond[i] = k

            if len(k) == 2:
                self.need_to_orient = True
                if self.ws.orientation.shape[:3] != self.ws.matrix.shape or \
                        not 0.9 < np.min(np.linalg.norm(self.ws.orientation[(self.ws.matrix >= low) &
                                                                            (self.ws.matrix <= high)], axis=1)) < 1.1:
                    raise Exception("The Workspace needs an orientation in order to align the conductivities.")

            # segmenting tmp domain to check if all values covered by mat_cond
            ws_tmp_tocheck[(self.ws.matrix >= low) & (self.ws.matrix <= high)] = i

        unique_matrixvalues = np.unique(ws_tmp_tocheck)
        if (unique_matrixvalues.size != self.cond_map.get_size() or
                np.all(np.array(list(self.mat_cond.keys()), dtype=np.uint16) != unique_matrixvalues)):
            raise Exception("All values in workspace must be included in ConductivityMap.")

        # print_matrices checks
        if type(self.print_matrices) is not tuple or len(self.print_matrices) != 5:
            raise Exception("Print_matrices must be a tuple with 5 booleans.")
        return False

    # Printing functions of system matrices
    def _print_E(self, i, i_cv, dec=4):
        np.set_printoptions(precision=dec)
        for j in range(self.len_y - 1):
            for k in range(self.len_z - 1):
                print()
                print("index {},{},{}".format(i_cv, j, k))
                print(self.Emat[i, j, k])

    def _print_A(self, dec=4):
        np.set_printoptions(linewidth=10000)
        np.set_printoptions(threshold=sys.maxsize)
        np.set_printoptions(formatter={'float': lambda x: "{:.{}f}".format(x, dec).rstrip('0').rstrip('.')})
        print()
        print("A matrix:")
        print(self.Amat.toarray())

    def _print_b(self, dec=1):
        print()
        print("b vector:")
        for k in range(self.len_z):
            for i in range(self.len_x):
                for j in range(self.len_y):
                    print('{:.{}f}'.format(self.bvec[self.len_x * (self.len_y * k + j) + i, 0], dec), end=' ')
                print()
            print()


def print_T(temperature, dec=4):
    print()
    print("Temperature:")
    for k in range(temperature.shape[2]):
        for i in range(temperature.shape[0]):
            for j in range(temperature.shape[1]):
                print('{:.{}f}'.format(temperature[i, j, k], dec), end=' ')
            print()
        print()


def print_flux(flux, dec=4):
    print()
    print("Flux:")
    for k in range(flux.shape[2]):
        for i in range(flux.shape[0]):
            for j in range(flux.shape[1]):
                print('({:.{}f}, {:.{}f}, {:.{}f})'.format(flux[i, j, k, 0], dec,
                                                           flux[i, j, k, 1], dec,
                                                           flux[i, j, k, 2], dec), end=' ')
            print()
        print()



import numpy as np


def get_sphere(diameter):
    radius = diameter / 2.
    shape = (int(diameter+4), int(diameter+4), int(diameter+4))
    center_point = (diameter+4.) / 2. - 0.5
    point = np.array([center_point, center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)-point[:, None, None, None], axis=0)
    sphere = (255.-127.5*(1+np.tanh((dist-radius))))
    sphere[sphere < 0] = 0
    sphere = sphere.astype(np.uint16)
    return sphere


def get_circle(diameter):
    radius = diameter / 2.
    shape = (int(diameter+4), int(diameter+4))
    center_point = (int(diameter+4)) / 2.
    point = np.array([center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)+0.5-point[:, None, None], axis=0)
    circle = (255.-128*(1+np.tanh(2*(dist-radius))))
    circle[circle < 0] = 0
    circle = circle.astype(np.uint16)
    return circle


def get_circle_odd(diameter):
    radius = diameter / 2.
    if int(diameter) % 2 == 0:
        shape = (int(diameter+5), int(diameter+5))
        center_point = (int(diameter + 5)) / 2.
    else:
        shape = (int(diameter+4), int(diameter+4))
        center_point = (int(diameter + 4)) / 2.
    point = np.array([center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)+0.5-point[:, None, None], axis=0)
    circle = (255.-128*(1+np.tanh(2*(dist-radius))))
    circle[circle < 0] = 0
    circle = circle.astype(np.uint16)
    return circle, shape[0]


def get_circle_even(diameter):
    radius = diameter / 2.
    if int(diameter) % 2 == 0:
        shape = (int(diameter+4), int(diameter+4))
        center_point = (int(diameter + 4)) / 2.
    else:
        shape = (int(diameter+5), int(diameter+5))
        center_point = (int(diameter + 5)) / 2.
    point = np.array([center_point, center_point])
    dist = np.linalg.norm(np.indices(shape)+0.5-point[:, None, None], axis=0)
    circle = (255.-128*(1+np.tanh(2*(dist-radius))))
    circle[circle < 0] = 0
    circle = circle.astype(np.uint16)
    return circle, shape[0]



from pumapy.utilities.logger import Logger, print_warning
import skimage.transform as trans
import numpy as np
from copy import deepcopy
from scipy.ndimage import rotate


class Workspace:
    def __init__(self, **kwargs):
        """ Workspace class holding the domain as a numpy matrix

        :param kwargs: keyword arguments
        - No arguments --> zero matrix of shape (1,1,1)
        - 'shape' argument --> then set all to zeros
        - 'value' --> value set all to full of value
        - 'nparray' --> set Workspace matrix as input array
        - 'scalars' --> bool indicating if a Workspace contains scalars stored in matrix variable
        - 'vectors' --> bool indicating if a Workspace contains vectors stored in orientation variable
        :type kwargs: dict
        """
        self.log = Logger()
        self.voxel_length = 1e-6

        # 3D numpy array holding the geometry (scalar)
        self.matrix = np.zeros((1, 1, 1), dtype=np.uint16)

        # 3D numpy array holding the orientation field (vectors)
        self.orientation = np.zeros((1, 1, 1, 3), dtype=float)

        # setting matrix
        if len(kwargs) > 0:
            if 'shape' in kwargs:
                if isinstance(kwargs['shape'], tuple) and len(kwargs['shape']) == 3:
                    if 'value' in kwargs:
                        self.matrix = np.full(kwargs['shape'], kwargs['value'], dtype=np.uint16)
                    else:
                        self.matrix = np.zeros(kwargs['shape'], dtype=np.uint16)
                else:
                    raise Exception("Wrong shape, tuple with 3 dimensions required.")

            elif 'nparray' in kwargs:
                if isinstance(kwargs['nparray'], np.ndarray):
                    if kwargs['nparray'].ndim == 3:
                        self.matrix = kwargs['nparray'].copy().astype(np.uint16)
                    else:
                        raise Exception("Wrong nparray ndim, 3 dimensions required.")
                else:
                    raise Exception("Wrong nparray type.")
            else:
                raise Exception("Unrecognized keyword.")

        # setting orientation
        if 'vectors' in kwargs and kwargs['vectors']:
            if not isinstance(kwargs['vectors'], bool):
                raise Exception("Orientation input is not a bool.")
            else:
                if 'shape' in kwargs:
                    if isinstance(kwargs['shape'], tuple) and len(kwargs['shape']) == 3:
                        if 'vectorvalue' in kwargs:
                            self.orientation = np.tile(kwargs['vectorvalue'], list(kwargs['shape']) + [1]).astype(float)
                        else:
                            self.orientation = np.zeros(list(kwargs['shape']) + [3], dtype=float)
                    else:
                        raise Exception("Wrong shape, tuple with 3 dimensions required.")
                else:
                    raise Exception("Unrecognized keyword.")

    @classmethod
    def from_shape(cls, shape, orientation=False):
        """ Generate workspace from shape, all matrix value set to zero.

        :param shape: shape of workspace to be created
        :type shape: tuple(int, int, int)
        :param orientation: specify if workspace contains orientation
        :type orientation: bool, optional
        :return: new workspace
        :rtype: Workspace
        """
        return cls(shape=shape, vectors=orientation)

    @classmethod
    def from_shape_value(cls, shape, value, orientation=False):
        """ Generate workspace from shape, all matrix values set to the value passed.

        :param shape: shape of workspace to be created
        :type shape: tuple(int, int, int)
        :param value: value to be assigned to the matrix variable
        :type value: int
        :param orientation: specify if workspace contains orientation
        :type orientation: bool, optional
        :return: new workspace
        :rtype: Workspace
        """
        return cls(shape=shape, value=value, vectors=orientation)

    @classmethod
    def from_shape_vector(cls, shape, vector):
        """ Generate workspace from shape, all orientation vectors set to the vector passed.

        :param shape: shape of workspace to be created
        :type shape: tuple(int, int, int)
        :param vector: vector to be assigned to the orientation variable
        :type vector: tuple(float, float, float)
        :return: new workspace with orientation
        :rtype: Workspace
        """
        return cls(shape=shape, vectorvalue=vector, vectors=True)

    @classmethod
    def from_shape_value_vector(cls, shape, value, vector):
        """ Generate workspace from shape, all matrix and orientation set to the values passed.

        :param shape: shape of workspace to be created
        :type shape: tuple(int, int, int)
        :param value: value to be assigned to the matrix variable
        :type value: int
        :param vector: vector to be assigned to the orientation variable
        :type vector: tuple(float, float, float)
        :return: new workspace with orientation
        :rtype: Workspace
        """
        return cls(shape=shape, value=value, vectorvalue=vector, vectors=True)

    @classmethod
    def from_array(cls, nparray):
        """ Generate workspace matrix from numpy array.

        :param nparray: array of shape (X,Y,Z) to be assigned to the matrix variable
        :type nparray: ndarray
        :return: new workspace
        :rtype: Workspace
        """
        return cls(nparray=nparray, vectors=False)

    def set_voxel_length(self, voxel_length):
        """ Set voxel size, which by default is set to 1e-6

        :param voxel_length: size of a voxel side
        :type voxel_length: float
        :return: None
        """
        if not isinstance(voxel_length, int) and not isinstance(voxel_length, float):
            raise Exception("Voxel_length needs to be an int or float, got " + str(type(voxel_length)))
        else:
            self.voxel_length = voxel_length

    def set_matrix(self, nparray):
        """ Set matrix with numpy array

        :param nparray: array of shape (X,Y,Z) to be assigned to the matrix variable
        :type nparray: ndarray
        :return: None
        """
        if isinstance(nparray, np.ndarray):
            if nparray.ndim == 3:
                self.matrix = nparray.copy().astype(np.uint16)
            else:
                raise Exception("Wrong nparray ndim, 3 dimensions required. Leaving matrix unchanged")
        else:
            print_warning("Wrong nparray type. Leaving matrix unchanged")

    def set_orientation(self, nparray):
        """ Set orientation with numpy array

        :param nparray: array of shape (X,Y,Z, 3) to be assigned to the orientation variable
        :type nparray: ndarray
        :return: None
        """
        if isinstance(nparray, np.ndarray):
            if nparray.ndim == 4 and nparray.shape[3] == 3:
                self.orientation = nparray.copy().astype(float)
            else:
                raise Exception("Wrong nparray ndim, 4 dimensions required as (x,y,z,3). Leaving orientation unchanged")
        else:
            raise Exception("Wrong nparray type. Leaving orientation unchanged")

    def copy(self):
        """ Create a copy of the workspace

        :return: copy of workspace
        :rtype: Workspace
        """
        return deepcopy(self)

    def __getitem__(self, key):
        return self.matrix[key]

    def __setitem__(self, key, value):
        self.matrix[key] = value

    def get_size(self):
        return self.matrix.size

    def len_x(self):
        return self.matrix.shape[0]

    def len_y(self):
        return self.matrix.shape[1]

    def len_z(self):
        return self.matrix.shape[2]

    def ndim(self):
        return self.matrix.ndim

    def get_shape(self):
        return self.matrix.shape

    def min(self):
        return np.min(self.matrix)

    def max(self):
        return np.max(self.matrix)

    def average(self):
        return self.matrix.mean()

    def unique_values(self):
        return np.unique(self.matrix)

    def unique_values_counts(self):
        return np.unique(self.matrix, return_counts=True)

    def orientation_magnitude(self):
        return np.linalg.norm(self.orientation, axis=3)

    def resize_new_matrix(self, shape, value=None):
        """ Resize matrix numpy array

        :param shape: shape of workspace to be resized
        :type shape: tuple(int, int, int)
        :param value: value to be assigned to the new resized matrix variable
        :type value: int, optional
        """
        if isinstance(shape, tuple) and len(shape) == 3:
            if value is None:
                self.matrix = np.zeros(shape, dtype=np.uint16)
            else:
                self.matrix = np.full(shape, value, dtype=np.uint16)
        else:
            raise Exception("Wrong shape, tuple with 3 dimensions required.")

    def resize_new_orientation(self, shape, orientation_value=None):
        """ Resize orientation numpy array

        :param shape: shape of workspace to be resized
        :type shape: tuple(int, int, int)
        :param orientation_value: vector to be assigned to the new resized orientation variable
        :type orientation_value: tuple(float, float, float), optional
        """
        if isinstance(shape, tuple) and len(shape) == 3:
            if orientation_value is None:
                self.orientation = np.zeros(list(shape) + [3], dtype=float)
            else:
                if isinstance(orientation_value, tuple) and len(orientation_value) == 3:
                    self.orientation = np.tile(orientation_value, list(shape) + [1]).astype(float)
                else:
                    raise Exception("vectorvalue has to be tuple(float, float, float).")
        else:
            raise Exception("Wrong shape, tuple with 3 dimensions required.")

    def create_orientation(self):
        """ Create orientation field of the same size as the matrix 

        :return: None
        """
        self.orientation = np.zeros(list(self.matrix.shape) + [3], dtype=float)

    def resize(self, shape, segmented, anti_aliasing=True, interpolation_order=1):
        """ Resize both matrix and orientation (if present) by rescaling the content to specified size

        :param shape: shape of workspace to be resized
        :type shape: tuple(int, int, int)
        :param segmented: specifying whether the domain is already segmented (True) or grayscales (False)
        :type segmented: bool
        :param anti_aliasing: if aliasing is to be prevented applying a Gaussian filter to smooth
            before scaling. If domain is segmented, automatically set to False in order to preserve domain
        :type anti_aliasing: bool, optional
        :param interpolation_order: order of the interpolation spline used.
            For segmented, it is enforced to be 0,which is 'nearest neighbor' to preserve the segmentation
        :type interpolation_order: int, optional
        :return: None
        """
        if isinstance(shape, tuple) and len(shape) == 3:
            if self.orientation.shape[:3] == self.matrix.shape:
                self.orientation = trans.resize(self.orientation, list(shape) + [3], order=0,
                                                preserve_range=True, anti_aliasing=False)
            if segmented:
                self.matrix = trans.resize(self.matrix, shape, order=0, anti_aliasing=False, preserve_range=True)
            else:
                self.matrix = trans.resize(self.matrix, shape, order=interpolation_order,
                                           anti_aliasing=anti_aliasing, preserve_range=True)
            self.matrix = self.matrix.astype('uint16')
        else:
            raise Exception("Wrong shape, tuple with 3 dimensions required.")

    def rescale(self, scale, segmented, anti_aliasing=True, interpolation_order=1):
        """ Rescale both matrix and orientation (if present) by rescaling the content by a specified factor

        :param scale: specifying the scaling factor
        :type scale: float
        :param segmented: specifying whether the domain is already segmented (True) or grayscales (False)
        :type segmented: bool
        :param anti_aliasing: if aliasing is to be prevented applying a Gaussian filter
            to smooth before scaling. If domain is segmented, automatically set to False in order to preserve domain
        :type anti_aliasing: bool, optional
        :param interpolation_order: order of the interpolation spline used.
            For segmented, it is enforced to be 0, which is 'nearest neighbor' to preserve the segmentation
        :type interpolation_order: int, optional
        :return: None
        """

        unit_dim_check = None
        if min(self.get_shape()) == 1:
            if self.len_x() == 1:
                unit_dim_check = 0
            elif self.len_y() == 1:
                unit_dim_check = 1
            elif self.len_z() == 1:
                unit_dim_check = 2
            self.matrix = np.squeeze(self.matrix)

        if self.orientation.shape[:3] == self.matrix.shape:
            self.orientation = trans.rescale(self.orientation, scale, order=0, multichannel=True,
                                             preserve_range=True, anti_aliasing=False)
        if segmented:
            self.matrix = trans.rescale(self.matrix, scale, order=0, anti_aliasing=False, preserve_range=True)
        else:
            self.matrix = trans.rescale(self.matrix, scale, order=interpolation_order,
                                        anti_aliasing=anti_aliasing, preserve_range=True)

        if unit_dim_check is not None:
            self.matrix = np.expand_dims(self.matrix, axis=unit_dim_check)

        self.matrix = self.matrix.astype('uint16')
        print("Rescaled workspace size: {}".format(self.get_shape()))

    def set(self, matrix_value=None, orientation_value=None):
        """ Set all elements in matrix equal to value (and orientation to vectorvalue is passed)

        :param matrix_value: value to fill to the matrix variable
        :type matrix_value: np.uint16, optional
        :param orientation_value: vector to fill to the orientation variable
        :type orientation_value: (tuple(float, float, float), optional)
        :return: None
        """
        check = True
        if matrix_value is not None:
            if np.issubdtype(type(matrix_value), np.integer) and 0 <= matrix_value <= 65535:
                self.matrix.fill(np.uint16(matrix_value))
                check = False
            else:
                raise Exception("matrix_value has to be np.uint16 (i.e. 0 <= matrix_value <= 65535).")
        if orientation_value is not None:
            if isinstance(orientation_value, tuple) and len(orientation_value) == 3:
                if self.orientation.shape[:3] != self.matrix.shape:
                    self.resize_new_orientation(self.matrix.shape, orientation_value=orientation_value)
                else:
                    self.orientation[:] = orientation_value
                check = False
            else:
                raise Exception("orientation_value has to be tuple(float, float, float).")
        if check:
            print_warning("No changes have been made to the Workspace.")

    def apply_mask(self, mask, apply_to_orientation=False):
        """ Apply mask of same size as the matrix by leaving the mask's 1s unchanged and setting mask's 0s to 0

        :param mask: array of type bool with same size as matrix
        :type mask: ndarray
        :param apply_to_orientation: specifying whether the mask is to be applied to the orientation (if present)
        :type apply_to_orientation: bool, optional
        :return: None
        """
        if isinstance(mask, np.ndarray) and mask.dtype == 'bool':
            if mask.shape == self.matrix.shape:
                self.matrix[~mask] = 0
                if self.orientation.shape[:3] == self.matrix.shape and apply_to_orientation:
                    self.orientation[~mask] = 0
            else:
                raise Exception("The mask has to be of the same size as the Workspace matrix.")
        else:
            raise Exception("The mask has to be a Numpy array of type bool.")

    def set_material_id(self, cutoff, value):
        """ Threshold the workspace according to cutoff (i.e. tuple with inclusive range to set)

        :param cutoff: convert a range of grayscale values specified by the cutoff into an single ID number
        :type cutoff: tuple(int, int)
        :param value: ID number to assign to range
        :type value: int
        :return: None
        """
        if value < 0:
            raise Exception("Value ID can only be positive. Leaving matrix unchanged.")
        if value > 1000000:
            raise Exception("Value ID cannot be > 1000000. Leaving matrix unchanged.")

        self.matrix = np.where(np.logical_and(self.matrix >= cutoff[0], self.matrix <= cutoff[1]), np.uint16(value), self.matrix)

        self.log.log_section("Set Material ID")
        self.log.log_line(str(cutoff) + " -> " + str(value))
        self.log.write_log()

    def binarize(self, threshold):
        """ Binarize the workspace according to threshold, inclusive for higher range set to 1, lower to 0

        :param threshold: grayscale value dividing the domain into 0s and 1s (threshold turns into 1)
        :type threshold: int
        :return: None
        """
        self.matrix = np.where(self.matrix < threshold, np.uint16(0), np.uint16(1))

        self.log.log_section("Binarize domain")
        self.log.log_line(">" + str(threshold) + " -> 1; 0 otherwise")
        self.log.write_log()

    def binarize_range(self, ones_cutoff):
        """ Binarize the workspace according to range within cutoff, inclusive for cutoff ints set to 1, rest to 0

        :param ones_cutoff: convert a range of grayscale values specified by the cutoff into 1s, rest into 0s
        :type ones_cutoff: tuple(int, int)
        :return: None
        """
        self.matrix = np.where(np.logical_and(self.matrix >= ones_cutoff[0], self.matrix <= ones_cutoff[1]), np.uint16(1), np.uint16(0))

        self.log.log_section("Binarize domain")
        self.log.log_line(str(ones_cutoff[0]) + " < matrix < " + str(ones_cutoff[1]) + " -> 1; 0 otherwise")
        self.log.write_log()

    def rotate(self, degrees, around_axis, reshape=False, boundary_mode='reflect', apply_to_orientation=True):
        """ Rotate domain by specified degrees

        :param degrees: degrees to rotate domain
        :type degrees: float
        :param around_axis: specify around what axis to perform the rotation. It can be 'x', 'y' or 'z'
        :type around_axis: string
        :param reshape: specify whether to reshape the domain to contain every voxel or keep it as original size
        :type reshape: bool, optional
        :param boundary_mode: specifying what to do with the boundaries.
            Options: ‘reflect’, ‘constant’, ‘nearest’, ‘mirror’, ‘wrap’
        :type boundary_mode: string, optional
        :param apply_to_orientation: specify whether to apply rotation to the orientation, if present
        :type apply_to_orientation: bool, optional
        :return: None
        """
        if around_axis == 'x':
            axes = (1, 2)
        elif around_axis == 'y':
            axes = (0, 2)
        elif around_axis == 'z':
            axes = (0, 1)
        else:
            raise Exception("Axis not recognized: around_axis can only be 'x', 'y' or 'z'")

        self.matrix = rotate(self.matrix, angle=degrees, axes=axes, mode=boundary_mode, reshape=reshape)

        if self.orientation.shape[:3] == self.matrix.shape and apply_to_orientation:
            from scipy.spatial.transform import Rotation
            rotation_degrees = degrees
            rotation_radians = np.radians(rotation_degrees)
            if around_axis == 'x':
                rotation_axis = np.array([1, 0, 0])
            elif around_axis == 'y':
                rotation_axis = np.array([0, 1, 0])
            else:
                rotation_axis = np.array([0, 0, 1])
            rotation_vector = rotation_radians * rotation_axis
            rotation = Rotation.from_rotvec(rotation_vector)
            for i in range(self.len_x()):
                for j in range(self.len_y()):
                    for k in range(self.len_z()):
                        self.orientation[i, j, k] = rotation.apply(self.orientation[i, j, k])

            self.orientation = rotate(self.orientation, angle=degrees, axes=axes, mode=boundary_mode,
                                      reshape=reshape, order=0)

    def show_matrix(self):
        if isinstance(self, Workspace):
            x, y, z = self.matrix.shape
        elif isinstance(self, np.ndarray):
            x, y, z = self.shape
        else:
            raise Exception("Print can only be called on a Workspace or Numpy array.")
        print()
        print("3D Workspace:")
        # Printing coordinate system used
        print("  o---> y")
        print("  |")
        print("x v")
        print('[', end='')
        for k in range(z):
            print("(:,:,{})".format(k))
            print('[', end='')
            for i in range(x):
                print('[', end='')
                for j in range(y):
                    print(self[i, j, k], end='')
                    if j != y - 1:
                        print(' ', end='')
                print(']', end='')
                if i != x - 1:
                    print()
            if k != z - 1:
                print()
                print()
        print(']')

    def show_orientation(self, dec=1):
        if isinstance(self, Workspace):
            x, y, z = self.matrix.shape
        elif isinstance(self, np.ndarray):
            if self.ndim == 4 and self.shape[3] == 3:
                x, y, z, _ = self.shape
            else:
                raise Exception("Numpy array has to be of size (x,y,z,3).")
        else:
            raise Exception("Print can only be called on a Workspace or Numpy array.")
        print()
        print("3D Orientation:")
        # Printing coordinate system used
        print("  o---> y")
        print("  |")
        print("x v")
        print('[', end='')
        for k in range(z):
            print("(:,:,{})".format(k))
            print('[', end='')
            for i in range(x):
                print('[', end='')
                for j in range(y):
                    if isinstance(self, Workspace):
                        print('({:.{}f}, {:.{}f}, {:.{}f})'.format(self.orientation[i, j, k, 0], dec,
                                                                   self.orientation[i, j, k, 1], dec,
                                                                   self.orientation[i, j, k, 2], dec), end='')
                    else:
                        print('({:.{}f}, {:.{}f}, {:.{}f})'.format(self[i, j, k, 0], dec,
                                                                   self[i, j, k, 1], dec,
                                                                   self[i, j, k, 2], dec), end='')
                    if j != y - 1:
                        print(' ', end='')
                print(']', end='')
                if i != x - 1:
                    print()
            if k != z - 1:
                print()
                print()
        print(']')



from pumapy import Workspace
from pumapy.utilities.timer import Timer
from pumapy.generation.tpms_utils import generate
import numpy as np


def generate_tpms(size, w, q, equation=0):
    """ Generation of triply periodic minimal surface material

    :param size: size of 3D domain (x,y,z)
    :type size: tuple(int, int, int)
    :param w: w parameter for tpms
    :type w: float or tuple(float, float)
    :param q: q parameter for tpms (float or tuple)
    :type q: float or tuple(float, float)
    :param equation: equation 0, 1, or 2 for tpms
    :type equation: int
    :return: TPMS domain
    :rtype: Workspace

    :Example:
    >>> import pumapy as puma
    >>> size = (200, 200, 200)  # size of the domain, in voxels.
    >>> w = 0.08  # value of w in the equations above
    >>> q = 0.2  # value of q in the equations above
    >>> ws_eq0 = puma.generate_tpms(size, w, q, equation=0)
    >>> ws_eq1 = puma.generate_tpms(size, w, q, equation=1)
    >>> ws_eq2 = puma.generate_tpms(size, w, q, equation=2)
    >>> puma.render_contour(ws_eq0, cutoff=(128, 255)) #visualize the workspace
    >>> puma.render_contour(ws_eq1, cutoff=(128, 255))
    >>> puma.render_contour(ws_eq2, cutoff=(128, 255))
    """

    if isinstance(w, tuple):
        _w = w
    elif isinstance(w, float):
        _w = (w, w)
    else:
        raise Exception("Invalid w, must be float or tuple")

    if isinstance(q, tuple):
        _q = q
    elif isinstance(q, float):
        _q = (q, q)
    else:
        raise Exception("Invalid w, must be float or tuple")

    generator = GeneratorTPMS(size, _w, _q, equation)

    generator.error_check()

    generator.log_input()
    generator.log_output()

    return generator.generate()


class GeneratorTPMS:
    def __init__(self, size, w, q, equation=0):
        self._workspace = Workspace()
        self._size = size
        self._wmin = w[0]
        self._wmax = w[1]
        self._qmin = q[0]
        self._qmax = q[1]
        self._equation = equation

    def generate(self):
        t = Timer()
        self._workspace.resize_new_matrix(self._size)
        self._workspace.matrix = generate(self._workspace.matrix.astype(float), self._size[0], self._size[1],
                                          self._size[2], self._wmin, self._wmax, self._qmin, self._qmax, self._equation)
        self._workspace.matrix = (255. * (self._workspace.matrix - np.min(self._workspace.matrix)) /
                                  np.ptp(self._workspace.matrix)).astype(np.uint16)
        print("Generated in: " + str(t.elapsed()) + " seconds")
        return self._workspace

    def log_input(self):
        self._workspace.log.log_section("Generating TPMS")
        self._workspace.log.log_line("Domain Size: " + str(self._size))
        self._workspace.log.log_line("w: " + str(self._wmin) + ", " + str(self._wmax))
        self._workspace.log.log_line("w: " + str(self._qmin) + ", " + str(self._qmax))
        self._workspace.log.log_line("Equation: " + str(self._equation))
        self._workspace.log.write_log()

    def log_output(self):
        self._workspace.log.log_section("Finished TPMS Generation")
        self._workspace.log.write_log()

    def error_check(self):
        if not isinstance(self._workspace, Workspace):
            raise Exception("Error, workspace must be passed")
        if not (isinstance(self._size, tuple) and len(self._size) == 3):
            raise Exception("Error, invalid size, must be tuple of size 3")
        if self._size[0] <= 0:
            raise Exception("Error, invalid size, must be >= 0")
        if self._size[1] <= 0:
            raise Exception("Error, invalid size, must be >= 0")
        if self._size[2] <= 0:
            raise Exception("Error, invalid size, must be >= 0")
        if self._wmin <= 0:
            raise Exception("Error, invalid w, must be >= 0")
        if self._wmax <= 0:
            raise Exception("Error, invalid w, must be >= 0")
        if self._wmax < self._wmin:
            raise Exception("Error, invalid w, w[1] cannot be < w[0]")
        if self._qmin <= 0:
            raise Exception("Error, invalid q, must be >= 0")
        if self._qmax <= 0:
            raise Exception("Error, invalid q, must be >= 0")
        if self._qmax < self._wmin:
            raise Exception("Error, invalid q, q[1] cannot be < q[0]")
        if self._equation != 0 and self._equation != 1 and self._equation != 2:
            raise Exception("Error, invalid equation number. Must be 0, 1, or 2")



import unittest
import numpy as np
import pumapy as puma


class TestAnisotropicTC(unittest.TestCase):

    def test_analytical_direct(self):
        ws = puma.generate_2d_square_array(100, 1. - 2. * np.pi * (0.1 ** 2.))
        ws.binarize_range((128, 255))
        ws.voxel_length = 1. / ws.matrix.shape[0]
        keff, _, _, _ = puma.compute_permeability(ws, (1, 1), solver_type='direct')
        np.testing.assert_array_almost_equal(keff, np.array([[2.71223274e-02, 0., 0.], [0., 2.71223274e-02, 0.], [0., 0., 5.48134246e-02]]))

    def test_analytical_bicgstab(self):
        ws = puma.generate_2d_square_array(100, 1. - 2. * np.pi * (0.1 ** 2.))
        ws.binarize_range((128, 255))
        ws.voxel_length = 1. / ws.matrix.shape[0]
        keff, _, _, _ = puma.compute_permeability(ws, (1, 1), solver_type='bicgstab', tol=1e-7, maxiter=10000)
        np.testing.assert_array_almost_equal(keff, np.array([[2.71223274e-02, 0., 0.], [0., 2.71223274e-02, 0.], [0., 0., 5.48134246e-02]]), decimal=4)

    def test_analytical_cg(self):
        ws = puma.generate_2d_square_array(100, 1. - 2. * np.pi * (0.1 ** 2.))
        ws.binarize_range((128, 255))
        ws.voxel_length = 1. / ws.matrix.shape[0]
        keff, _, _, _ = puma.compute_permeability(ws, (1, 1), solver_type='cg', tol=1e-7, maxiter=10000)
        np.testing.assert_array_almost_equal(keff, np.array([[2.71223274e-02, 0., 0.], [0., 2.71223274e-02, 0.], [0., 0., 5.48134246e-02]]), decimal=4)

    def test_analytical_minres(self):
        ws = puma.generate_2d_square_array(100, 1. - 2. * np.pi * (0.1 ** 2.))
        ws.binarize_range((128, 255))
        ws.voxel_length = 1. / ws.matrix.shape[0]
        keff, _, _, _ = puma.compute_permeability(ws, (1, 1), solver_type='minres', tol=1e-8, maxiter=10000)
        np.testing.assert_array_almost_equal(keff, np.array([[2.71223274e-02, 0., 0.], [0., 2.71223274e-02, 0.], [0., 0., 5.48134246e-02]]), decimal=4)

if __name__ == '__main__':
    unittest.main()



from matplotlib import pyplot as plt
import numpy as np
import pumapy.utilities.workspace as ws


def plot_slices(ws_nparray, slice_direction='z', crange=None, cmap='gray', index=1):
    """ Plot slices of domain along a specified direction (z default)

    :param ws_nparray: domain
    :type ws_nparray: Workspace or ndarray
    :param slice_direction: 'x', 'y', 'z'
    :type slice_direction: string
    :param crange: color range, i.e. specify min and max grayscale
    :type crange: tuple(int, int)
    :param cmap: color map for the plot, 'gray' (default), 'jet' or refer to matplotlib for other colormaps
    :type cmap: string
    :param index: specifying the slice index in which the plot will be opened
    :type: int
    :return: slicer object
    :rtype: PlotSlicer

    >>> import pumapy as puma
    >>> ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
    >>> puma.plot_slices(ws)
    """

    img, _ = PlotSlicer.error_checks(ws_nparray, None, slice_direction)
    if img is None:
        return

    img, axis_labels, slices, rows, cols = PlotSlicer.rotate_domain_mpl(img, slice_direction)
    return PlotSlicer(img, slice_direction, crange, cmap, index, axis_labels, slices, rows, cols)


def compare_slices(ws_nparray1, ws_nparray2, slice_direction='z', crange1=None, cmap1='gray',
                   crange2=None, cmap2='gray', index=1):
    """ Plot slices of domain along a specified direction (z default)

    :param ws_nparray1: domain
    :type ws_nparray1: Workspace or ndarray
    :param ws_nparray2: domain
    :type ws_nparray2: Workspace or ndarray
    :param slice_direction: 'x', 'y', 'z'
    :type slice_direction: string
    :param crange1: color range for plot 1, specify min and max grayscale
    :type crange1: tuple(int, int)
    :param cmap1: color map for plot 1, 'gray' (default), 'jet' or refer to matplotlib for other colormaps
    :type cmap1: string
    :param crange2: color range for plot 2, specify min and max grayscale
    :type crange2: tuple(int, int)
    :param cmap2: color map for plot 2, 'gray' (default), 'jet' or refer to matplotlib for other colormaps
    :type cmap2: string
    :param index: specifying the slice index in which the plot will be opened
    :type index: int
    :return: slicer object
    :rtype: CompareSlicer

    >>> import pumapy as puma
    >>> ws = puma.import_3Dtiff(puma.path_to_example_file("100_fiberform.tif"), 1.3e-6)
    >>> ws2 = ws.copy()
    >>> ws2.binarize_range((100, 255))
    >>> puma.compare_slices(ws, ws2)
    """

    img1, img2 = CompareSlicer.error_checks(ws_nparray1, ws_nparray2, slice_direction)
    if img1 is None:
        return

    img1, axis_labels, slices1, rows1, cols1 = CompareSlicer.rotate_domain_mpl(img1, slice_direction)
    img2, _, slices2, rows2, cols2 = CompareSlicer.rotate_domain_mpl(img2, slice_direction)
    return CompareSlicer(img1, img2, slice_direction, crange1, cmap1, crange2, cmap2, index,
                         axis_labels, slices1, slices2, rows1, cols1, rows2, cols2)


class IndexTracker:
    def __init__(self, img, img2, slice_direction, index, axis_labels, slices, slices_titles):
        self.slice_direction = slice_direction
        self.ind = index - 1
        self.axis_labels = axis_labels
        self.slices = slices
        self.slices_titles = slices_titles
        self.axis_labels = axis_labels
        self.img = img
        self.img2 = img2

    def onkey(self, event):
        if event.key == "right":
            if self.ind < self.slices - 10:
                self.ind += 10
        elif event.key == "left":
            if self.ind >= 10:
                self.ind -= 10
        elif event.key == "up":
            if self.ind < self.slices - 100:
                self.ind += 100
        elif event.key == "down":
            if self.ind >= 100:
                self.ind -= 100
        self.update()

    def onscroll(self, event):
        if event.button == 'down':
            if self.ind != self.slices - 1:
                self.ind += 1
        else:
            if self.ind != 0:
                self.ind -= 1
        self.update()

    def update(self):
        pass  # virtual

    @staticmethod
    def format_coord(x, y):
        return 'x=%1.f, y=%1.f' % (x, y)

    @staticmethod
    def rotate_domain_mpl(img, slice_direction):
        if slice_direction == 'x':
            img = np.swapaxes(img, 0, 2)
            img = np.swapaxes(img, 0, 1)
            slices, rows, cols = img.shape
            axis_labels = ('y', 'z')
        elif slice_direction == 'y':
            img = np.swapaxes(img, 1, 2)
            img = np.swapaxes(img, 1, 0)
            rows, slices, cols = img.shape
            axis_labels = ('z', 'x')
        else:
            rows, cols, slices = img.shape
            axis_labels = ('x', 'y')

        img = np.rot90(img, axes=(1, 0))
        img = np.flip(img, axis=1)
        return img, axis_labels, slices, rows, cols

    @staticmethod
    def _ws_check(img):
        if isinstance(img, ws.Workspace):
            img = img.matrix.copy()
        elif isinstance(img, np.ndarray):
            if img.ndim > 3 or img.ndim < 2:
                raise Exception("Numpy array has to have 3 dimensions")
            elif img.ndim == 2:
                img = img[:, :, np.newaxis]
            img = img.copy()
        else:
            raise Exception("Data to be sliced has to be either a pumapy.Workspace or np.ndarray")
        return img

    @staticmethod
    def error_checks(img, img2, slice_direction):
        slice_direction = slice_direction.lower()
        if not (slice_direction == 'x' or slice_direction == 'y' or slice_direction == 'z'):
            raise Exception("Slice direction can only be along 'x', 'y' or 'z'")

        img = PlotSlicer._ws_check(img)
        if img is None:
            return None, None

        if img2 is not None:
            img2 = PlotSlicer._ws_check(img2)
            if img2 is None:
                return None, None
            if not (slice_direction == 'x' or slice_direction == 'y' or slice_direction == 'z'):
                raise Exception("Slice direction can only be along 'x', 'y' or 'z'")
        return img, img2


class PlotSlicer(IndexTracker):
    def __init__(self, img, slice_direction, color_range, color_map, index, axis_labels, slices, rows, cols):
        super().__init__(img, None, slice_direction, index, axis_labels, slices, None)
        self.color_range = color_range
        self.color_map = color_map
        self.rows = rows
        self.cols = cols

        if self.color_range is None:
            self.color_range = self.img.min(), self.img.max()

        self.fig, self.ax = plt.subplots(1, 1)
        self.im = self.ax.imshow(self.img[:, :, self.ind], cmap=self.color_map,
                                 vmin=self.color_range[0], vmax=self.color_range[1])

        self.cid_scroll = self.fig.canvas.mpl_connect('scroll_event', self.onscroll)
        self.cid_key = self.fig.canvas.mpl_connect('key_press_event', self.onkey)

        self.ax.set_xlabel(self.axis_labels[0])
        self.ax.set_ylabel(self.axis_labels[1])
        plt.colorbar(mappable=self.im, ax=self.ax, fraction=0.046, pad=0.04,
                     ticks=np.linspace(self.color_range[0], self.color_range[1], 15).astype(int))
        self.ax.grid(linestyle=':')
        self.update()
        plt.show(block=True)

    def update(self):
        self.im.set_data(self.img[:, :, self.ind])
        self.ax.set_title('Slice: {}/{} along {}'.format(self.ind + 1, self.slices, self.slice_direction))
        self.ax.format_coord = PlotSlicer.format_coord
        self.fig.canvas.draw()


class CompareSlicer(IndexTracker):
    def __init__(self, img1, img2, slice_direction, color_range1, color_map1, color_range2, color_map2, index,
                 axis_labels, slices1, slices2, rows1, cols1, rows2, cols2):
        super().__init__(img1, img2, slice_direction, index, axis_labels, max(slices1, slices2), [slices1, slices2])
        self.imgs = img1, img2
        self.color_ranges = [color_range1, color_range2]
        self.color_maps = color_map1, color_map2
        self.rows = rows1, rows2
        self.cols = cols1, cols2

        self.fig, self.ax = plt.subplots(1, 2)
        self.fig.tight_layout(pad=3.0)
        self.ims = [None, None]

        self.cid_scroll = self.fig.canvas.mpl_connect('scroll_event', self.onscroll)
        self.cid_key = self.fig.canvas.mpl_connect('key_press_event', self.onkey)

        for i in range(2):
            if self.color_ranges[i] is None:
                self.color_ranges[i] = (self.imgs[i].min(), self.imgs[i].max())

            self.ims[i] = self.ax[i].imshow(self.imgs[i][:, :, self.ind], cmap=self.color_maps[i],
                                            vmin=self.color_ranges[i][0], vmax=self.color_ranges[i][1])
            self.ax[i].set_xlabel(self.axis_labels[0])
            self.ax[i].set_ylabel(self.axis_labels[1])
            plt.colorbar(self.ims[i], ax=self.ax[i], fraction=0.046, pad=0.04,
                         ticks=np.linspace(self.color_ranges[i][0], self.color_ranges[i][1], 15).astype(int))
            self.ax[i].grid(linestyle=':')
        self.update()
        plt.show(block=True)

    def update(self):
        for i in range(2):
            self.ims[i].set_data(self.imgs[i][:, :, self.ind])
            self.ax[i].set_title('Slice: {}/{} along {}'.format(self.ind + 1, self.slices_titles[i], self.slice_direction))
            self.ax[i].format_coord = CompareSlicer.format_coord
        self.fig.canvas.draw()



from pumapy import Workspace
from pumapy.utilities.timer import Timer
import pumapy.utilities.generic_checks as check
from pumapy.generation.sphere import get_circle_even, get_circle_odd
import numpy as np


def generate_2d_square_array(size, porosity):
    """ Generate a 2D periodic array of circles

    :param size: size of the output domain
    :type size: int
    :param porosity: porosity of the output domain
    :type porosity: float
    :return: array of circles
    :rtype: Workspace
    """
    generator = GeneratorSquareArray(size, porosity)

    generator.error_check()

    generator.log_input()
    generator.log_output()

    return generator.generate()


class GeneratorSquareArray:
    def __init__(self, size, porosity):
        self._workspace = Workspace()
        self._size = (size, size, 1)
        self._center = (np.array([size, size]) / 2) + np.array([size, size])
        self._porosity = porosity

        self._diameter = np.sqrt((1-porosity) * size*size / (2 * np.pi)) * 2
        self._size_padded = (self._size[0] * 3, self._size[1]*3, 1)

    def generate(self):
        t = Timer()
        self._workspace.resize_new_matrix(self._size_padded)
        self._generate_circle()
        print("Generated in: " + str(t.elapsed()) + " seconds")
        return self._workspace

    def _generate_circle(self):

        if self._size[0] % 2 == 0:
            circle, circlesize = get_circle_even(self._diameter)
        else:
            circle, circlesize = get_circle_odd(self._diameter)

        self._circlesize = circlesize
        if self._circlesize % 2 == 0:
            self._size_m = self._circlesize / 2
            self._size_p = self._circlesize / 2
        else:
            self._size_m = int(self._circlesize / 2.)
            self._size_p = int(self._circlesize / 2.) + 1

        [x,y] = self._center
        xm = int(x - self._size_m)
        xp = int(x + self._size_p)
        ym = int(y - self._size_m)
        yp = int(y + self._size_p)
        self._workspace.matrix[xm:xp, ym:yp] += circle[:,:,None]

        circle, circlesize = get_circle_even(self._diameter)
        self._circlesize = circlesize
        if self._circlesize % 2 == 0:
            self._size_m = self._circlesize / 2
            self._size_p = self._circlesize / 2
        else:
            self._size_m = int(self._circlesize / 2.)
            self._size_p = int(self._circlesize / 2.) + 1

        xvec = [self._size[0], self._size[0] * 2]
        yvec = [self._size[1], self._size[1] * 2]

        for x in xvec:
            for y in yvec:
                xm = int(x - self._size_m)
                xp = int(x + self._size_p)
                ym = int(y - self._size_m)
                yp = int(y + self._size_p)
                self._workspace.matrix[xm:xp, ym:yp] += circle[:, :, None]

        self._workspace.matrix = self._workspace.matrix[self._size[0]:2 * self._size[0], self._size[1]:2 * self._size[1]]

    def log_input(self):
        self._workspace.log.log_section("Generating Square Array")
        self._workspace.log.log_value("Domain Size", self._size)
        self._workspace.log.log_value("Porosity", self._porosity)
        self._workspace.log.log_value("Diameter", self._diameter)
        self._workspace.log.write_log()

    def log_output(self):
        self._workspace.log.log_section("Finished Generating Square Array")
        self._workspace.log.write_log()

    def error_check(self):
        if not check.size_check(self._size):
            return False
        if self._porosity < 0.214602:
            raise Exception("Error, invalid porosity. Cylinders cannot fit with a porosity of < 0.214602")



import numpy as np
from skimage import measure
import scipy.ndimage as ndimage
import pyvista as pv
from pumapy.utilities.workspace import Workspace
from pumapy.utilities.generic_checks import check_ws_cutoff


def generate_isosurface(workspace, cutoff, flag_closed_edges=True, flag_gaussian=False):
    """ Generation of isosurface based on cutoff provided

    :param workspace: domain
    :type workspace: Workspace or numpy.ndarray
    :param cutoff: specify cutoff to binarize material
    :type cutoff: tuple(int, int)
    :param flag_closed_edges: close edges of surfaces on the boundaries
    :type flag_closed_edges: bool, optional
    :param flag_gaussian: apply Gaussian filter before generating surface
    :type flag_gaussian: bool, optional
    :return: triangulated surface
    :rtype: TriMesh

    :Example:
    >>> import pumapy as puma
    >>> ws_isosurface = puma.import_3Dtiff(puma.path_to_example_file("200_fiberform.tif"))
    >>> ws_copy = ws_isosurface.copy()
    >>> puma.utilities.isosurface.generate_isosurface(ws_copy, (128, 255))
    >>> puma.compare_slices(ws_copy, ws_isosurface)
    """
    iso = Isosurface(workspace, cutoff, flag_closed_edges, flag_gaussian)

    iso.error_check()

    iso.log_input()
    iso.compute()
    iso.log_output()

    return iso.tri_mesh


class TriMesh:
    def __init__(self):
        self.verts = None
        self.faces = None
        self.normals = None
        self.values = None
        self.mesh = None

    def copy(self, other):
        self.verts = np.copy(other.verts)
        self.faces = np.copy(other.faces)
        self.normals = np.copy(other.normals)
        self.values = np.copy(other.values)

    def create_mesh(self):
        f = np.zeros((self.faces.shape[0], 4), dtype=np.uint32)
        f[:, 0] = 3
        f[:, 1:] = self.faces
        self.mesh = pv.PolyData(self.verts, f.ravel())

    def save(self, filename, binary=True):
        if not filename.lower().endswith('.stl'):
            filename += ".stl"
        if self.mesh is None:
            self.create_mesh()
        self.mesh.save(filename, binary=binary)


class Isosurface:
    def __init__(self, workspace, cutoff, flag_closed_edges, flag_gaussian):
        if isinstance(workspace, np.ndarray):
            self.workspace = Workspace.from_array(workspace)
        else:
            self.workspace = workspace
        self.cutoff = list(cutoff)
        self.flag_closed_edges = flag_closed_edges
        self.flag_gaussian = flag_gaussian
        self.tri_mesh = TriMesh()
        self.matrix = np.empty(self.workspace.matrix.shape)

    def compute(self):
        if float(self.cutoff[0]).is_integer():
            self.cutoff[0] -= 0.5
        if float(self.cutoff[1]).is_integer():
            self.cutoff[1] += 0.5

        self.flip_matrix()
        self.apply_gaussian()
        self.buffer_matrix()
        import warnings
        warnings.simplefilter(action='ignore', category=FutureWarning)
        self.tri_mesh.verts, self.tri_mesh.faces, self.tri_mesh.normals, self.tri_mesh.values \
            = measure.marching_cubes_lewiner(self.matrix, self.iso_value)

    def flip_matrix(self):
        self.matrix = np.copy(self.workspace.matrix).astype(np.float32)
        average = (self.cutoff[0] + self.cutoff[1]) / 2.0
        mask = self.matrix > average
        self.matrix[mask] = 2 * average - self.matrix[mask]
        self.iso_value = self.cutoff[0]

    def buffer_matrix(self):
        if self.flag_closed_edges:
            matrix_buf = np.full([self.matrix.shape[0] + 2, self.matrix.shape[1] + 2, self.matrix.shape[2] + 2], -1e3)
            matrix_buf[1:-1, 1:-1, 1:-1] = self.matrix[:, :, :]
            self.matrix = np.copy(matrix_buf)

    def apply_gaussian(self):
        if self.flag_gaussian:
            self.matrix = ndimage.gaussian_filter(self.matrix, sigma=1)

    def error_check(self):
        check_ws_cutoff(self.workspace, self.cutoff)

    def log_input(self):
        self.workspace.log.log_section("Computing Isosurface")
        self.workspace.log.log_line("Domain Size: " + str(self.workspace.get_shape))
        self.workspace.log.log_line("Cutoff: " + str(self.cutoff))

        if self.flag_closed_edges:
            self.workspace.log.log_line("Closed Edges: True")
        else:
            self.workspace.log.log_line("Closed Edges: False")

        if self.flag_gaussian:
            self.workspace.log.log_line("Gaussian: True")
        else:
            self.workspace.log.log_line("Gaussian: False")

        self.workspace.log.write_log()

    def log_output(self):
        self.workspace.log.log_section("Finished Isosurface")
        self.workspace.log.log_line("Number of Triangles: " + str(self.tri_mesh.values.shape[0] / 3.))



import unittest
import numpy as np
import pumapy as puma
from skimage import io


class TestSurfacearea(unittest.TestCase):

    def test_fiberform(self):
        ws = puma.Workspace.from_array(io.imread(puma.path_to_example_file("200_fiberform.tif")))
        area, specific_area = puma.compute_surface_area(ws, (90, 255), False)

        np.testing.assert_almost_equal(area, 4.12090e-7, decimal=5)
        np.testing.assert_almost_equal(specific_area, 52291.746, decimal=3)


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import pumapy as puma
from pumapy.physicsmodels.mpsa_elasticity import Elasticity
import scipy.sparse


class TestElasticity(unittest.TestCase):

    def setUp(self):
        self.X = 6
        self.Y = 4
        self.Z = 8

        self.ws_homoiso = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.elast_map_homoiso = puma.ElasticityMap()
        self.elast_map_homoiso.add_material((1, 1),
                                            1, 0.1, 0.2, 0.3, 0.4, 0.5,
                                            2, 0.6, 0.7, 0.8, 0.9,
                                            3, 0.11, 0.12, 0.13,
                                            4, 0.14, 0.15,
                                            5, 0.16,
                                            6)
        # elasticity solution tensor for homogeneous
        #    1       0.1     0.2     0.3     0.4     0.5
        #    0.1       2     0.6     0.7     0.8     0.9
        #    0.2     0.6     3       0.11    0.12    0.13
        #    0.3     0.7     0.11    4       0.14    0.15
        #    0.4     0.8     0.12    0.14    5       0.16
        #    0.5     0.9     0.13    0.15    0.16    6

        self.solution_homoiso_x = np.zeros((self.X, self.Y, self.Z), dtype=float)
        for i in range(self.X):
            self.solution_homoiso_x[i] = i / (self.X - 1.)

        self.solution_homoiso_y = np.zeros((self.X, self.Y, self.Z), dtype=float)
        for j in range(self.Y):
            self.solution_homoiso_y[:, j] = j / (self.Y - 1.)

        self.solution_homoiso_z = np.zeros((self.X, self.Y, self.Z), dtype=float)
        for k in range(self.Z):
            self.solution_homoiso_z[:, :, k] = k / (self.Z - 1.)

        self.ws_matSeriesInx = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.ws_matSeriesInx[int(self.ws_matSeriesInx.matrix.shape[0] / 2.):, :] = 2  # in series in x

        self.ws_matSeriesIny = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.ws_matSeriesIny[:, int(self.ws_matSeriesIny.matrix.shape[1] / 2.):] = 2  # in series in y

        self.ws_matSeriesInz = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        self.ws_matSeriesInz[:, :, int(self.ws_matSeriesInz.matrix.shape[2] / 2.):] = 2  # in series in z

        self.elast_matSeries = puma.ElasticityMap()
        self.elast_matSeries.add_material((1, 1),
                                          10, 0.2, 0.3, 0, 0, 0,
                                          10, 0.5, 0, 0, 0,
                                          10, 0, 0, 0,
                                          0.5, 0, 0,
                                          0.5, 0,
                                          0.5)
        self.elast_matSeries.add_material((2, 2),
                                          1, 0.2, 0.3, 0, 0, 0,
                                          1, 0.5, 0, 0, 0,
                                          1, 0, 0, 0,
                                          0.5, 0, 0,
                                          0.5, 0,
                                          0.5)
        # elasticity solution tensor for mat in series along x
        # 1.8181 0.2 0.3
        #    0.2 5.5 0.5
        #    0.3 0.5 5.5

        # elasticity solution tensor for mat in series along y
        # 5.5    0.2 0.3
        # 0.2 1.8181 0.5
        # 0.3    0.5 5.5

        # elasticity solution tensor for mat in series along z
        # 5.5 0.2    0.3
        # 0.2 5.5    0.5
        # 0.3 0.5 1.8181

    def test_homoiso_x(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_homoiso, self.elast_map_homoiso, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [1, 0.1, 0.2, 0.3, 0.4, 0.5])
        np.testing.assert_array_almost_equal(u[:, :, :, 0], self.solution_homoiso_x)

    def test_homoiso_y(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_homoiso, self.elast_map_homoiso, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.1, 2, 0.6, 0.7, 0.8, 0.9])
        np.testing.assert_array_almost_equal(u[:, :, :, 1], self.solution_homoiso_y)

    def test_homoiso_z(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_homoiso, self.elast_map_homoiso, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.2, 0.6, 3, 0.11, 0.12, 0.13])
        np.testing.assert_array_almost_equal(u[:, :, :, 2], self.solution_homoiso_z)

    def test_matSeriesInx_x_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [1.818181818, 0.2, 0.3, 0, 0, 0])

    def test_matSeriesInx_y_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.2, 5.5, 0.5, 0, 0, 0])

    def test_matSeriesInx_z_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.3, 0.5, 5.5, 0, 0, 0])

    def test_matSeriesIny_x_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesIny, self.elast_matSeries, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [5.5, 0.2, 0.3, 0, 0, 0])

    def test_matSeriesIny_y_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesIny, self.elast_matSeries, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.2, 1.818181818, 0.5, 0, 0, 0])

    def test_matSeriesIny_z_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesIny, self.elast_matSeries, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.3, 0.5, 5.5, 0, 0, 0])

    def test_matSeriesInz_x_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInz, self.elast_matSeries, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [5.5, 0.2, 0.3, 0, 0, 0])

    def test_matSeriesInz_y_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInz, self.elast_matSeries, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.2, 5.5, 0.5, 0, 0, 0])

    def test_matSeriesInz_z_per(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInz, self.elast_matSeries, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.3, 0.5, 1.818181818, 0, 0, 0])

    def test_matSeriesInx_x_sym(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'x', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [1.818181818, 0.2, 0.3, 0, 0, 0])

    def test_matSeriesInx_y_sym(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'y', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.2, 5.5, 0.5, 0, 0, 0])

    def test_matSeriesInx_z_sym(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'z', 's', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [0.3, 0.5, 5.5, 0, 0, 0])

    def test_matSeriesInx_x_sym_bicgstab(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'x', 's', solver_type='bicgstab')
        np.testing.assert_array_almost_equal(Ceff, [1.818181818, 0.2, 0.3, 0, 0, 0], decimal=4)

    def test_matSeriesInx_y_sym_bicgstab(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'y', 's', solver_type='bicgstab')
        np.testing.assert_array_almost_equal(Ceff, [0.2, 5.5, 0.5, 0, 0, 0], decimal=4)

    def test_matSeriesInx_z_sym_bicgstab(self):
        Ceff, u, _, _ = puma.compute_elasticity(self.ws_matSeriesInx, self.elast_matSeries, 'z', 's', solver_type='bicgstab')
        np.testing.assert_array_almost_equal(Ceff, [0.3, 0.5, 5.5, 0, 0, 0], decimal=4)

    def test_symmetry(self):
        X, Y, Z = (8, 6, 4)
        ws = puma.Workspace.from_array(np.ones((X, Y, Z)))
        bc = puma.ElasticityBC.from_workspace(ws)

        elast_map = puma.ElasticityMap()
        elast_map.add_isotropic_material((1, 1), 10, 0.3)
        elast_map.add_isotropic_material((2, 2), 7.3, 0.23)

        # Along x
        bc[0, :, :, 0] = -1.
        bc[-1, :, :, 0] = 1.
        bc[0, :, :, [1, 2]] = 0.
        bc[-1, :, :, [1, 2]] = 0.
        # puma.Workspace.show_orientation(bc)

        ws[:, :int(Y/2)] = 2
        # ws.show_matrix()

        u, _, _ = puma.compute_stress_analysis(ws, elast_map, bc, side_bc='p', solver_type='direct')
        # puma.Workspace.show_orientation(u, 5)

        np.testing.assert_array_almost_equal(u[:int(X / 2), :, :, 0], -u[int(X / 2):, :, :, 0][::-1], decimal=4)
        np.testing.assert_array_almost_equal(u[:int(X / 2), :, :, 1],  u[int(X / 2):, :, :, 1][::-1], decimal=4)
        np.testing.assert_array_almost_equal(u[:int(X / 2), :, :, 2],  u[int(X / 2):, :, :, 2][::-1], decimal=4)

        np.testing.assert_array_almost_equal(np.array([[0.14197716, 0.04368197, 0.],
                                                       [0.42584737, 0.04142283, 0.],
                                                       [0.71028350, 0.03215080, 0.],
                                                       [1.00000000, 0.00000000, 0.]], dtype=float), u[int(X / 2):, 2, 2], decimal=7)

        # Along y
        bc = puma.ElasticityBC.from_workspace(ws)
        bc[:, 0, :, 1] = -1.
        bc[:, -1, :, 1] = 1.
        bc[:, 0, :, [0, 2]] = 0.
        bc[:, -1, :, [0, 2]] = 0.
        # puma.Workspace.show_orientation(bc)

        ws = puma.Workspace.from_array(np.ones((X, Y, Z)))
        ws[:int(X/2)] = 2
        # ws.show_matrix()

        u, _, _ = puma.compute_stress_analysis(ws, elast_map, bc, side_bc='p', solver_type='direct')
        # puma.Workspace.show_orientation(u, 5)

        np.testing.assert_array_almost_equal(u[:, :int(Y / 2), :, 0],  u[:, int(Y / 2):, :, 0][:, ::-1], decimal=4)
        np.testing.assert_array_almost_equal(u[:, :int(Y / 2), :, 1], -u[:, int(Y / 2):, :, 1][:, ::-1], decimal=4)
        np.testing.assert_array_almost_equal(u[:, :int(Y / 2), :, 2],  u[:, int(Y / 2):, :, 2][:, ::-1], decimal=4)

        # Along z
        bc = puma.ElasticityBC.from_workspace(ws)
        bc[:, :, 0, 2] = -1.
        bc[:, :, -1, 2] = 1.
        bc[:, :, 0, [0, 1]] = 0.
        bc[:, :, -1, [0, 1]] = 0.
        # puma.Workspace.show_orientation(bc)

        u, _, _ = puma.compute_stress_analysis(ws, elast_map, bc, side_bc='p', solver_type='direct')
        # puma.Workspace.show_orientation(u, 5)

        np.testing.assert_array_almost_equal(u[:, :, :int(Z / 2), 0],  u[:, :, int(Z / 2):, 0][:, :, ::-1], decimal=4)
        np.testing.assert_array_almost_equal(u[:, :, :int(Z / 2), 1],  u[:, :, int(Z / 2):, 1][:, :, ::-1], decimal=4)
        np.testing.assert_array_almost_equal(u[:, :, :int(Z / 2), 2], -u[:, :, int(Z / 2):, 2][:, :, ::-1], decimal=4)

    def test_tensor_rotation_x(self):
        ws = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        ws.set(orientation_value=(1, 0, 0))
        elast_map = puma.ElasticityMap()
        elast_map.add_material_to_orient((1, 1), 10, 20, 0.23, 0.3, 50)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [14.33251433, 9.41850942, 9.41850942, 0, 0, 0], decimal=7)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [9.41850942, 28.16732817, 12.78271278, 0, 0, 0], decimal=7)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [9.41850942, 12.78271278, 28.16732817, 0, 0, 0], decimal=7)

    def test_tensor_rotation_y(self):
        ws = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        ws.set(orientation_value=(0, 1, 0))
        elast_map = puma.ElasticityMap()
        elast_map.add_material_to_orient((1, 1), 10, 20, 0.23, 0.3, 50)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [28.167328167328126, 9.418509418509416, 12.782712782712776, 0, 0, 0], decimal=7)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [9.418509418509423, 14.33251433251433, 9.418509418509421, 0, 0, 0], decimal=7)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [12.78271278271278, 9.418509418509414, 28.16732816732822, 0, 0, 0], decimal=7)

    def test_tensor_rotation_z(self):
        ws = puma.Workspace.from_array(np.ones((self.X, self.Y, self.Z)))
        ws.set(orientation_value=(0, 0, 1))
        elast_map = puma.ElasticityMap()
        elast_map.add_material_to_orient((1, 1), 10, 20, 0.23, 0.3, 50)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'x', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [28.167328167328208, 12.782712782712784, 9.418509418509416, 0, 0, 0], decimal=7)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'y', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [12.782712782712778, 28.167328167328137, 9.418509418509416, 0, 0, 0], decimal=7)

        Ceff, u, _, _ = puma.compute_elasticity(ws, elast_map, 'z', 'p', solver_type='direct')
        np.testing.assert_array_almost_equal(Ceff, [9.418509418509425, 9.418509418509425, 14.33251433251433, 0, 0, 0], decimal=7)

    def test_Amat_builtinbeam596(self):
        ws = puma.Workspace.from_shape_value((5, 9, 6), 1)

        elast_map = puma.ElasticityMap()
        elast_map.add_isotropic_material((1, 1), 200, 0.3)

        bc = puma.ElasticityBC.from_workspace(ws)
        bc[0] = 0
        bc[-1] = [0, 1, 0]

        solver = Elasticity(ws, elast_map, None, 'f', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_xf.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        solver = Elasticity(ws, elast_map, None, 'p', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_xp.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        solver = Elasticity(ws, elast_map, None, 's', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_xs.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        bc = puma.ElasticityBC.from_workspace(ws)
        bc[:, 0] = 0
        bc[:, -1] = [0, 0, 1]

        solver = Elasticity(ws, elast_map, None, 'f', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_yf.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        solver = Elasticity(ws, elast_map, None, 'p', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_yp.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        solver = Elasticity(ws, elast_map, None, 's', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_ys.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        bc = puma.ElasticityBC.from_workspace(ws)
        bc[:, :, 0] = 0
        bc[:, :, -1] = [1, 0, 0]

        solver = Elasticity(ws, elast_map, None, 'f', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_zf.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        solver = Elasticity(ws, elast_map, None, 'p', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_zp.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)

        solver = Elasticity(ws, elast_map, None, 's', bc, None, None, "direct", True, (0, 0, 0, 0, 0))
        solver.error_check()
        solver.initialize()
        solver.assemble_bvector()
        solver.assemble_Amatrix()
        Amat_correct = scipy.sparse.load_npz('testdata/mpsa_Amat/Amat_builtinbeam596_zs.npz')
        test_Amat = np.abs(solver.Amat.toarray() - Amat_correct.toarray())
        self.assertAlmostEqual(test_Amat.max(), 0, 10)


if __name__ == '__main__':
    unittest.main()



from skimage.filters import gaussian
from scipy.ndimage import distance_transform_edt, median_filter
from scipy.ndimage.filters import uniform_filter
from skimage.morphology import erosion, dilation, opening, closing, ball
from pumapy import Workspace
from pumapy.utilities.logger import print_warning
import numpy as np


def filter_median(ws, size):
    """ 3D Median filter

    :param ws: input workspace
    :type ws: Workspace
    :param size: size of window
    :type size: int
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws.binarize_range((128, 255))  # binarize it so that it is only 1s and 0s
    >>> ws_median = ws.copy()
    >>> puma.filter_median(ws_median, size = 10)
    >>> puma.compare_slices(ws, ws_median) # compare it

    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed.")

    ws.log.log_section("Median Filter:")
    ws.log.log_line("Window Size: " + str(size))
    ws.log.write_log()

    ws.matrix = median_filter(ws.matrix, size)
    return True


def filter_gaussian(ws, sigma=1, apply_on_orientation=False):
    """ 3D Gaussian filter

        :param ws: input workspace
        :type ws: Workspace
        :param sigma: size of kernel
        :type sigma: int
        :param apply_on_orientation: specify whether to apply filter on orientation field
        :type apply_on_orientation: bool, optional
        :return: True if successful, False otherwise.
        :rtype: bool

        :Example:
        >>> import pumapy as puma
        >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0) #  generate tpms material
        >>> ws.binarize_range((128, 255)) # binarize it so that it is only 1s and 0s
        >>> ws_gaussian = ws.copy()
        >>> puma.filter_gaussian(ws_gaussian, sigma=2, apply_on_orientation=False)
        >>> puma.compare_slices(ws, ws_gaussian) # compare it
    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed.")

    ws.log.log_section("Gaussian Filter:")
    ws.log.log_line("Sigma: " + str(sigma))
    ws.log.write_log()

    ws.matrix = gaussian(ws.matrix, sigma=sigma, preserve_range=True).astype(np.uint16)

    if apply_on_orientation:
        if ws.orientation.shape[:3] == ws.matrix.shape:
            unprocessed_orientation = ws.orientation.copy()
            ws.orientation = gaussian(ws.orientation, sigma=sigma, multichannel=True, preserve_range=True)
            # recreate unit vectors
            magnitude = np.linalg.norm(ws.orientation, axis=3)
            with np.errstate(divide='ignore', invalid='ignore'):  # temporarily ignore divide by 0 warnings
                ws.orientation = ws.orientation / magnitude[:, :, :, np.newaxis]
            ws.orientation[np.isnan(ws.orientation)] = unprocessed_orientation[np.isnan(ws.orientation)]
        else:
            print_warning("To apply filter on orientation field, it has to have same shape as ws.matrix.")
    return True


def filter_edt(ws, cutoff):
    """ 3D Exact euclidean distance transform.

    :param ws: input workspace
    :type ws: Workspace
    :param cutoff: cutoff to binarize image
    :type cutoff: tuple(int, int)
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws.binarize_range((128, 255))  # binarize it so that it is only 1s and 0s
    >>> ws_edt = ws.copy()
    >>> puma.filter_edt(ws_edt, cutoff=(1,1))
    >>> puma.compare_slices(ws, ws_edt) #compare it
    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed")

    ws.log.log_section("Euclidean Distance Transform Filter")
    ws.log.write_log()

    ws.matrix = distance_transform_edt(np.logical_and(ws.matrix >= cutoff[0], ws.matrix <= cutoff[1]))
    return True


def filter_mean(ws, size=5):
    """ 3D Mean filter.

    :param ws: input workspace
    :type ws: Workspace
    :param size: size of window
    :type size: int, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws.binarize_range((128, 255))  # binarize it so that it is only 1s and 0s
    >>> ws_mean = ws.copy()
    >>> puma.filter_mean(ws_mean, size=10)
    >>> puma.compare_slices(ws, ws_mean) #compare it

    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed")

    ws.log.log_section("Mean Filter")
    ws.log.log_line("Window Size: " + str(size))
    ws.log.write_log()

    ws.matrix = uniform_filter(ws.matrix, size)
    return True


def filter_erode(ws, cutoff, size=5):
    """ 3D morphological erosion filter.

    :param ws: input workspace
    :type ws: Workspace
    :param cutoff: cutoff to binarize image
    :type cutoff: tuple(int, int)
    :param size: size of the spherical windows used
    :type size: int, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws.binarize_range((128, 255))  # binarize it so that it is only 1s and 0s
    >>> ws_erode = ws.copy()
    >>> puma.filter_erode(ws_erode, (1, 1))  # eroding the copy
    >>> puma.compare_slices(ws, ws_erode)  # compare it
    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed")

    ws.log.log_section("Erode Filter")
    ws.log.log_line("Spherical Window Size: " + str(size))
    ws.log.write_log()

    ws.binarize_range(cutoff)

    ws.matrix = erosion(ws.matrix, ball(size))
    return True


def filter_dilate(ws, cutoff, size=5):
    """ 3D morphological dilation filter.

    :param ws: input workspace
    :type ws: Workspace
    :param cutoff: cutoff to binarize image
    :type cutoff: tuple(int, int)
    :param size: size of the spherical windows used
    :type size: int, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws.binarize_range((128, 255))  # binarize it so that it is only 1s and 0s
    >>> ws_dilate = ws.copy()
    >>> puma.filter_dilate(ws_dilate, cutoff=(1, 1), size=5)  # dilating the copy
    >>> puma.compare_slices(ws, ws_dilate)  # compare it
    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed")

    ws.log.log_section("Dilation Filter")
    ws.log.log_line("Spherical Window Size: " + str(size))
    ws.log.write_log()

    ws.binarize_range(cutoff)

    ws.matrix = dilation(ws.matrix, ball(size))
    return True


def filter_opening(ws, cutoff, size=5):
    """ 3D morphological opening filter (i.e. dilation first and then erosion).

    :param ws: input workspace
    :type ws: Workspace
    :param cutoff: cutoff to binarize image
    :type cutoff: tuple(int, int)
    :param size: size of the spherical windows used
    :type size: int, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws_opening = ws.copy()
    >>> puma.filter_opening(ws_opening, cutoff=(128,255), size=3)
    >>> ws_binary = ws.copy()
    >>> ws_binary.binarize_range((128, 255))
    >>> puma.compare_slices(ws_binary, ws_opening) # compare it
    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed")

    ws.log.log_section("Opening Filter")
    ws.log.log_line("Spherical Window Size: " + str(size))
    ws.log.write_log()

    ws.binarize_range(cutoff)

    ws.matrix = opening(ws.matrix, ball(size))
    return True


def filter_closing(ws, cutoff, size=5):
    """ 3D morphological closing filter (i.e. erosion first and then dilation).

    :param ws: input workspace
    :type ws: Workspace
    :param cutoff: cutoff to binarize image
    :type cutoff: tuple(int, int)
    :param size: size of the spherical windows used
    :type size: int, optional
    :return: True if successful, False otherwise.
    :rtype: bool

    :Example:
    >>> import pumapy as puma
    >>> ws = puma.generate_tpms((100, 100, 100), (0.02, 0.05), 0.201, 0)  # generate tpms material
    >>> ws_closing = ws.copy()
    >>> puma.filter_closing(ws.closing, cutoff=(128, 255), size=3)
    >>> ws_binary = ws.copy()
    >>> ws_binary.binarize_range((128, 255))
    >>> puma.compare_slices(ws_binary, ws_closing) # compare it
    """
    if not isinstance(ws, Workspace):
        raise Exception("Workspace must be passed")

    ws.log.log_section("Closing Filter")
    ws.log.log_line("Spherical Window Size: " + str(size))
    ws.log.write_log()

    ws.binarize_range(cutoff)

    ws.matrix = closing(ws.matrix, ball(size))
    return True



from pumapy.physicsmodels.fe_permeability import Permeability


def compute_permeability(workspace, solid_cutoff, tol=1e-8, maxiter=10000, solver_type='minres', display_iter=True):
    """ Compute the permeability using first order Q1-Q1 Finite Element solver and periodic BC on the sides

    :param workspace: domain
    :type workspace: Workspace
    :param solid_cutoff: specify the solid phase
    :type solid_cutoff: tuple(int, int)
    :param tol: tolerance for iterative solver
    :type tol: float, optional
    :param maxiter: maximum Iterations for solver
    :type maxiter: int, optional
    :param solver_type: solver type, options: 'minres' (default), 'direct', 'cg', 'bicgstab'
    :type solver_type: string, optional
    :type display_iter: bool, optional
    :return: permeability, velocity field
    :rtype: tuple(ndarray, ndarray)
    """
    solver = Permeability(workspace, solid_cutoff, tol, maxiter, solver_type, display_iter)

    solver.error_check()

    solver.log_input()
    solver.compute()
    solver.log_output()
    return solver.keff, solver.u_x, solver.u_y, solver.u_z
